__author__ = "sibirrer"


import numpy as np
import pytest
import numpy.testing as npt
from lenstronomy.Util import param_util


class TestCSP(object):
    """Tests the cored steep ellipsoid (CSE)"""

    def setup_method(self):
        from lenstronomy.LensModel.Profiles.cored_steep_ellipsoid import CSE

        self.CSP = CSE(axis="product_avg")

    def test_function(self):
        kwargs = {"a": 2, "s": 1, "e1": 0.0, "e2": 0.0, "center_x": 0, "center_y": 0}

        x = np.array([1.0, 2])
        y = np.array([2, 0])
        f_ = self.CSP.function(x, y, **kwargs)
        npt.assert_almost_equal(f_, [1.09016, 0.96242], decimal=5)

    def test_derivatives(self):
        kwargs = {"a": 2, "s": 1, "e1": 0.0, "e2": 0.0, "center_x": 0, "center_y": 0}

        x = np.array([1.0, 2])
        y = np.array([2, 0])
        f_x, f_y = self.CSP.derivatives(x, y, **kwargs)
        npt.assert_almost_equal(f_x, [0.2367, 0.55279], decimal=5)
        npt.assert_almost_equal(f_y, [0.4734, 0.0], decimal=5)

    def test_hessian(self):
        kwargs = {"a": 2, "s": 1, "e1": 0.0, "e2": 0.0, "center_x": 0, "center_y": 0}

        x = np.array([1.0, 2])
        y = np.array([2, 0])
        f_xx, f_xy, f_yx, f_yy = self.CSP.hessian(x, y, **kwargs)
        npt.assert_almost_equal(f_xy, f_yx, decimal=5)
        npt.assert_almost_equal(f_xx, [0.16924, -0.09751], decimal=5)
        npt.assert_almost_equal(f_xy, [-0.13493, -0.0], decimal=5)
        npt.assert_almost_equal(f_yy, [-0.03315, 0.27639], decimal=5)

    def test_ellipticity(self):
        """Test the definition of the ellipticity normalization (along major axis or
        product averaged axes)"""
        x, y = np.linspace(start=0.001, stop=10, num=100), np.zeros(100)
        kwargs_round = {
            "a": 2,
            "s": 1,
            "e1": 0.0,
            "e2": 0.0,
            "center_x": 0,
            "center_y": 0,
        }
        phi_q, q = param_util.ellipticity2phi_q(0.3, 0)
        kwargs = {"a": 2, "s": 1, "e1": 0.3, "e2": 0.0, "center_x": 0, "center_y": 0}

        f_xx, f_xy, f_yx, f_yy = self.CSP.hessian(x, y, **kwargs_round)
        kappa_round = 1.0 / 2 * (f_xx + f_yy)

        f_xx, f_xy, f_yx, f_yy = self.CSP.hessian(x, y, **kwargs)
        kappa_major = 1.0 / 2 * (f_xx + f_yy)

        f_xx, f_xy, f_yx, f_yy = self.CSP.hessian(y, x, **kwargs)
        kappa_minor = 1.0 / 2 * (f_xx + f_yy)

        # import matplotlib.pyplot as plt
        # plt.plot(x, kappa_major/kappa_round, ',-', label='major/round', alpha=0.5)
        # plt.plot(x, kappa_minor/kappa_round, '--', label='minor/round', alpha=0.5)
        #
        # plt.plot(x, np.sqrt(kappa_minor*kappa_major)/kappa_round,label='prod/kappa_round')
        # plt.legend()
        # plt.show()

        npt.assert_almost_equal(
            kappa_round, np.sqrt(kappa_minor * kappa_major), decimal=1
        )


if __name__ == "__main__":
    pytest.main()



import numpy as np
import lenstronomy.Util.util as util
from lenstronomy.Data.pixel_grid import PixelGrid

__all__ = ["KinBin"]


class KinBin(object):
    """Class that summarizes the binned kinematic data.

    The KinBin() class is initialized with :
     - The information about the bins (bin values, and bin covariances, which pixels belong to which bin):
    bin_data, bin_cov, bin_mask.
     - The information about the associated intial shape of the unbinned kinematic map: bin_mask gives the index of
    corresponding bin for each pixel), and ra_at_xy_0,dec_at_xy_0,transform_pix2angle,ra_shift,dec_shift are the usual
    PixelGrid characteritics.
    """

    def __init__(
        self,
        bin_data,
        bin_cov,
        bin_mask,
        ra_at_xy_0,
        dec_at_xy_0,
        transform_pix2angle,
        psf_class,
        ra_shift=0,
        dec_shift=0,
    ):
        """
        :param bin_data: list, kinematic value of each bin, ordered by bin index.
        :param bin_cov: 2D array (nbins x nbins), vrms covariance matrix associated to each bin, ordered by bin index
        :param bin_mask: 2D array, mapping from the unbinned image to the binned one, each pixel value is the
         corresponding bin index.
        :param ra_at_xy_0: float, ra coordinate at pixel (0,0) (unbinned image)
        :param dec_at_xy_0: float, dec coordinate at pixel (0,0) (unbinned image)
        :param transform_pix2angle: 2x2 array, mapping of pixel (unbinned image) to coordinate
        :param psf_class: PSF class
        :param ra_shift:  float, RA shift of pixel grid
        :param dec_shift: float, DEC shift of pixel grid

        """
        self.PSF = psf_class
        nx, ny = np.shape(bin_mask)
        self._nx = nx
        self._ny = ny
        self.PixelGrid = PixelGrid(
            nx, ny, transform_pix2angle, ra_at_xy_0 + ra_shift, dec_at_xy_0 + dec_shift
        )

        self.data = bin_data
        self.covariance = bin_cov
        self.bin_mask = bin_mask
        self._pix2a = transform_pix2angle
        self._ra_at_xy_0 = ra_at_xy_0
        self._dec_at_xy_0 = dec_at_xy_0

    @staticmethod
    def binned_image(data, bin_mask):
        """Creates the binned image of the data.

        :param data: data value in each bin
        :param bin_mask: mask indicating which pixels belong to which bin
        """
        binned_image = np.zeros_like(bin_mask)
        for idx, value in enumerate(data):
            binned_image[bin_mask == idx] = value
        return binned_image

    def kin_bin2kwargs(self):
        """Creates the kwargs needed for the 2D kinematic likelihood."""
        kwargs = {
            "image": self.binned_image(self.data, self.bin_mask),
            "deltaPix": self.PixelGrid.pixel_width,
            "transform_pix2angle": self._pix2a,
            "ra_at_xy0": self._ra_at_xy_0,
            "dec_at_xy0": self._dec_at_xy_0,
        }
        return kwargs

    def kin_grid(self):
        """Creates a pixel grid that satisfy the kinematics coordinates system."""
        x_grid, y_grid = self.PixelGrid.pixel_coordinates
        return x_grid, y_grid



__author__ = "sibirrer"

from lenstronomy.LensModel.Profiles.sis_truncate import SIS_truncate

import numpy as np
import numpy.testing as npt
import pytest


class TestSIS_truncate(object):
    """Tests the Gaussian methods."""

    def setup_method(self):
        self.SIS = SIS_truncate()

    def test_function(self):
        x = 1
        y = 0
        phi_E = 1.0
        r_trunc = 2
        values = self.SIS.function(x, y, phi_E, r_trunc)
        assert values == 1
        x = np.array([0])
        y = np.array([0])
        phi_E = 1.0
        values = self.SIS.function(x, y, phi_E, r_trunc)
        assert values[0] == 0

        x = np.array([2, 3, 4])
        y = np.array([1, 1, 1])
        values = self.SIS.function(x, y, phi_E, r_trunc)
        npt.assert_almost_equal(values[0], 2.2221359549995796, decimal=9)
        npt.assert_almost_equal(values[1], 2.8245553203367586, decimal=9)
        assert values[2] == 3

    def test_derivatives(self):
        x = 1
        y = 2
        phi_E = 1.0
        r_trunc = 2
        f_x, f_y = self.SIS.derivatives(x, y, phi_E, r_trunc)
        npt.assert_almost_equal(f_x, 0.39442719099991586, decimal=9)
        npt.assert_almost_equal(f_y, 0.78885438199983171, decimal=9)
        x = np.array([0])
        y = np.array([0])
        f_x, f_y = self.SIS.derivatives(x, y, phi_E, r_trunc)
        assert f_x[0] == 0
        assert f_y[0] == 0

        x = np.array([1, 3, 4])
        y = np.array([2, 1, 1])
        values = self.SIS.derivatives(x, y, phi_E, r_trunc)
        npt.assert_almost_equal(values[0][0], 0.39442719099991586, decimal=9)
        npt.assert_almost_equal(values[1][0], 0.78885438199983171, decimal=9)
        npt.assert_almost_equal(values[0][1], 0.39736659610102748, decimal=9)
        npt.assert_almost_equal(values[1][1], 0.13245553203367583, decimal=9)

    def test_hessian(self):
        x = 1
        y = 0
        phi_E = 1.0
        r_trunc = 2
        f_xx, f_xy, f_yx, f_yy = self.SIS.hessian(x, y, phi_E, r_trunc)
        assert f_xx == 0
        assert f_yy == 1
        assert f_xy == 0
        npt.assert_almost_equal(f_xy, f_yx, decimal=8)
        x = np.array([1, 3, 4])
        y = np.array([2, 1, 1])
        values = self.SIS.hessian(x, y, phi_E, r_trunc)
        npt.assert_almost_equal(values[0][0], 0.21554175279993265, decimal=9)
        npt.assert_almost_equal(values[3][0], -0.3211145618000168, decimal=9)
        npt.assert_almost_equal(values[1][0], -0.3577708763999663, decimal=9)
        npt.assert_almost_equal(values[0][1], -0.43675444679663239, decimal=9)
        npt.assert_almost_equal(values[3][1], 0.06920997883030823, decimal=9)
        npt.assert_almost_equal(values[1][1], -0.18973665961010272, decimal=9)


if __name__ == "__main__":
    pytest.main()



import numpy as np
import lenstronomy.Util.param_util as param_util

from lenstronomy.Util.package_util import exporter

export, __all__ = exporter()


@export
class Gaussian(object):
    """Class for Gaussian light profile The two-dimensional Gaussian profile amplitude
    is defined such that the 2D integral leads to the 'amp' value.

    profile name in LightModel module: 'GAUSSIAN'
    """

    def __init__(self):
        self.param_names = ["amp", "sigma", "center_x", "center_y"]
        self.lower_limit_default = {
            "amp": 0,
            "sigma": 0,
            "center_x": -100,
            "center_y": -100,
        }
        self.upper_limit_default = {
            "amp": 1000,
            "sigma": 100,
            "center_x": 100,
            "center_y": 100,
        }

    def function(self, x, y, amp, sigma, center_x=0, center_y=0):
        """Surface brightness per angular unit.

        :param x: coordinate on the sky
        :param y: coordinate on the sky
        :param amp: amplitude, such that 2D integral leads to this value
        :param sigma: sigma of Gaussian in each direction
        :param center_x: center of profile
        :param center_y: center of profile
        :return: surface brightness at (x, y)
        """
        c = amp / (2 * np.pi * sigma**2)
        r2 = (x - center_x) ** 2 / sigma**2 + (y - center_y) ** 2 / sigma**2
        return c * np.exp(-r2 / 2.0)

    def total_flux(self, amp, sigma, center_x=0, center_y=0):
        """Integrated flux of the profile.

        :param amp: amplitude, such that 2D integral leads to this value
        :param sigma: sigma of Gaussian in each direction
        :param center_x: center of profile
        :param center_y: center of profile
        :return: total flux
        """
        return amp

    def light_3d(self, r, amp, sigma):
        """3D brightness per angular volume element.

        :param r: 3d distance from center of profile
        :param amp: amplitude, such that 2D integral leads to this value
        :param sigma: sigma of Gaussian in each direction
        :return: 3D brightness per angular volume element
        """
        amp3d = amp / np.sqrt(2 * sigma**2) / np.sqrt(np.pi)
        sigma3d = sigma
        return self.function(r, 0, amp3d, sigma3d)


@export
class GaussianEllipse(object):
    """Class for Gaussian light profile with ellipticity.

    profile name in LightModel module: 'GAUSSIAN_ELLIPSE'
    """

    param_names = ["amp", "sigma", "e1", "e2", "center_x", "center_y"]
    lower_limit_default = {
        "amp": 0,
        "sigma": 0,
        "e1": -0.5,
        "e2": -0.5,
        "center_x": -100,
        "center_y": -100,
    }
    upper_limit_default = {
        "amp": 1000,
        "sigma": 100,
        "e1": -0.5,
        "e2": -0.5,
        "center_x": 100,
        "center_y": 100,
    }

    def __init__(self):
        self.gaussian = Gaussian()

    def function(self, x, y, amp, sigma, e1, e2, center_x=0, center_y=0):
        """

        :param x: coordinate on the sky
        :param y: coordinate on the sky
        :param amp: amplitude, such that 2D integral leads to this value
        :param sigma: sigma of Gaussian in each direction
        :param e1: eccentricity modulus
        :param e2: eccentricity modulus
        :param center_x: center of profile
        :param center_y: center of profile
        :return: surface brightness at (x, y)
        """
        x_, y_ = param_util.transform_e1e2_product_average(
            x, y, e1, e2, center_x, center_y
        )
        return self.gaussian.function(x_, y_, amp, sigma, center_x=0, center_y=0)

    def total_flux(
        self, amp, sigma=None, e1=None, e2=None, center_x=None, center_y=None
    ):
        """Total integrated flux of profile.

        :param amp: amplitude, such that 2D integral leads to this value
        :param sigma: sigma of Gaussian in each direction
        :param e1: eccentricity modulus
        :param e2: eccentricity modulus
        :param center_x: center of profile
        :param center_y: center of profile
        :return: total flux
        """
        return self.gaussian.total_flux(amp, sigma, center_x, center_y)

    def light_3d(self, r, amp, sigma, e1=0, e2=0):
        """3D brightness per angular volume element.

        :param r: 3d distance from center of profile
        :param amp: amplitude, such that 2D integral leads to this value
        :param sigma: sigma of Gaussian in each direction
        :param e1: eccentricity modulus
        :param e2: eccentricity modulus
        :return: 3D brightness per angular volume element
        """
        return self.gaussian.light_3d(r, amp, sigma=sigma)


@export
class MultiGaussian(object):
    """Class for elliptical pseudo Jaffe lens light (2d projected light/mass
    distribution.

    profile name in LightModel module: 'MULTI_GAUSSIAN'
    """

    param_names = ["amp", "sigma", "center_x", "center_y"]
    lower_limit_default = {
        "amp": 0,
        "sigma": 0,
        "e1": -0.5,
        "e2": -0.5,
        "center_x": -100,
        "center_y": -100,
    }
    upper_limit_default = {
        "amp": 1000,
        "sigma": 100,
        "e1": -0.5,
        "e2": -0.5,
        "center_x": 100,
        "center_y": 100,
    }

    def __init__(self):
        self.gaussian = Gaussian()

    def function(self, x, y, amp, sigma, center_x=0, center_y=0):
        """Surface brightness per angular unit.

        :param x: coordinate on the sky
        :param y: coordinate on the sky
        :param amp: list of amplitudes of individual Gaussian profiles
        :param sigma: list of widths of individual Gaussian profiles
        :param center_x: center of profile
        :param center_y: center of profile
        :return: surface brightness at (x, y)
        """
        f_ = np.zeros_like(x)
        for i in range(len(amp)):
            f_ += self.gaussian.function(x, y, amp[i], sigma[i], center_x, center_y)
        return f_

    def total_flux(self, amp, sigma, center_x=0, center_y=0):
        """Total integrated flux of profile.

        :param amp: list of amplitudes of individual Gaussian profiles
        :param sigma: list of widths of individual Gaussian profiles
        :param center_x: center of profile
        :param center_y: center of profile
        :return: total flux
        """
        flux = 0
        for i in range(len(amp)):
            flux += self.gaussian.total_flux(amp[i], sigma[i], center_x, center_y)
        return flux

    def function_split(self, x, y, amp, sigma, center_x=0, center_y=0):
        """Split surface brightness in individual components.

        :param x: coordinate on the sky
        :param y: coordinate on the sky
        :param amp: list of amplitudes of individual Gaussian profiles
        :param sigma: list of widths of individual Gaussian profiles
        :param center_x: center of profile
        :param center_y: center of profile
        :return: list of arrays of surface brightness
        """
        f_list = []
        for i in range(len(amp)):
            f_list.append(
                self.gaussian.function(x, y, amp[i], sigma[i], center_x, center_y)
            )
        return f_list

    def light_3d(self, r, amp, sigma):
        """3D brightness per angular volume element.

        :param r: 3d distance from center of profile
        :param amp: list of amplitudes of individual Gaussian profiles
        :param sigma: list of widths of individual Gaussian profiles
        :return: 3D brightness per angular volume element
        """
        f_ = np.zeros_like(r)
        for i in range(len(amp)):
            f_ += self.gaussian.light_3d(r, amp[i], sigma[i])
        return f_


@export
class MultiGaussianEllipse(object):
    """Class for elliptical multi Gaussian profile.

    profile name in LightModel module: 'MULTI_GAUSSIAN_ELLIPSE'
    """

    param_names = ["amp", "sigma", "e1", "e2", "center_x", "center_y"]
    lower_limit_default = {
        "amp": 0,
        "sigma": 0,
        "e1": -0.5,
        "e2": -0.5,
        "center_x": -100,
        "center_y": -100,
    }
    upper_limit_default = {
        "amp": 1000,
        "sigma": 100,
        "e1": -0.5,
        "e2": -0.5,
        "center_x": 100,
        "center_y": 100,
    }

    def __init__(self):
        self.gaussian = Gaussian()

    def function(self, x, y, amp, sigma, e1, e2, center_x=0, center_y=0):
        """Surface brightness per angular unit.

        :param x: coordinate on the sky
        :param y: coordinate on the sky
        :param amp: list of amplitudes of individual Gaussian profiles
        :param sigma: list of widths of individual Gaussian profiles
        :param e1: eccentricity modulus
        :param e2: eccentricity modulus
        :param center_x: center of profile
        :param center_y: center of profile
        :return: surface brightness at (x, y)
        """
        x_, y_ = param_util.transform_e1e2_product_average(
            x, y, e1, e2, center_x, center_y
        )

        f_ = np.zeros_like(x)
        for i in range(len(amp)):
            f_ += self.gaussian.function(
                x_, y_, amp[i], sigma[i], center_x=0, center_y=0
            )
        return f_

    def total_flux(self, amp, sigma, e1, e2, center_x=0, center_y=0):
        """Total integrated flux of profile.

        :param amp: list of amplitudes of individual Gaussian profiles
        :param sigma: list of widths of individual Gaussian profiles
        :param e1: eccentricity modulus
        :param e2: eccentricity modulus
        :param center_x: center of profile
        :param center_y: center of profile
        :return: total flux
        """
        flux = 0
        for i in range(len(amp)):
            flux += self.gaussian.total_flux(amp[i], sigma[i], center_x, center_y)
        return flux

    def function_split(self, x, y, amp, sigma, e1, e2, center_x=0, center_y=0):
        """Split surface brightness in individual components.

        :param x: coordinate on the sky
        :param y: coordinate on the sky
        :param amp: list of amplitudes of individual Gaussian profiles
        :param sigma: list of widths of individual Gaussian profiles
        :param e1: eccentricity modulus
        :param e2: eccentricity modulus
        :param center_x: center of profile
        :param center_y: center of profile
        :return: list of arrays of surface brightness
        """
        x_, y_ = param_util.transform_e1e2_product_average(
            x, y, e1, e2, center_x, center_y
        )
        f_list = []
        for i in range(len(amp)):
            f_list.append(
                self.gaussian.function(x_, y_, amp[i], sigma[i], center_x=0, center_y=0)
            )
        return f_list

    def light_3d(self, r, amp, sigma, e1=0, e2=0):
        """3D brightness per angular volume element.

        :param r: 3d distance from center of profile
        :param amp: list of amplitudes of individual Gaussian profiles
        :param sigma: list of widths of individual Gaussian profiles
        :param e1: eccentricity modulus
        :param e2: eccentricity modulus
        :return: 3D brightness per angular volume element
        """
        f_ = np.zeros_like(r)
        for i in range(len(amp)):
            f_ += self.gaussian.light_3d(r, amp[i], sigma[i])
        return f_



import pytest
from lenstronomy.Conf import config_loader


def test_numba_conf():
    numba_conf = config_loader.numba_conf()
    assert "nopython" in numba_conf
    assert "cache" in numba_conf
    assert "parallel" in numba_conf
    assert "enable" in numba_conf
    assert "fastmath" in numba_conf
    assert "error_model" in numba_conf


def test_conventions_conf():
    conf = config_loader.conventions_conf()
    assert "sersic_major_axis" in conf


if __name__ == "__main__":
    pytest.main()



# import trained NN
import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable
import os
import json


class KinematicNN:
    """Class to call the NN to emulate JAM kinematics."""

    def __init__(self):
        current_directory = os.path.dirname(os.path.abspath(__file__))
        filepath = os.path.join(current_directory, "SKiNN_1.0_config.json")
        with open(filepath, "r") as json_file:
            self.config = json.load(json_file)
        try:
            import SKiNN
            from SKiNN.generator import Generator

            self.SKiNN_installed = True
            self.generator = Generator()
        except:
            print(
                "Warning : SKiNN not installed properly, \
        but tests will be trivially fulfilled. \
                  Get SKiNN from https://github.com/mattgomer/SKiNN"
            )
            self.SKiNN_installed = False

    def generate_map(self, input_p, verbose=False):
        """Generate velocity map given input parameters.

        :param input_p: vector of input parameters
        :return: the velocity map associated with input_p
        """
        self.within_bounds = self.check_bounds(input_p, verbose)
        return self.generator.generate_map(input_p)

    def plot_map(self, ax, input_p):
        self.check_bounds(input_p)
        im = ax.imshow(self.generate_map(input_p))
        ax.set_title("Prediction")
        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.05)
        cb = plt.colorbar(im, cax=cax)
        cb.set_label(r"$v_{\rm rms}$ [km/s]")
        return ax

    def check_bounds(self, input_p, same_orientation=True, verbose=False):
        """Checks to see if input parameters lie in bounds used for the training set.

        :param input_p: input parameters to NN
        :param same_orientation: default True; confirms that mass and light have same
            position angles
        :param verbose: default False; if True prints statements when out of bounds
        """
        within_bounds = True
        if not same_orientation:
            if verbose:
                print("NN CALL WARNING: Mass and light have different PAs!")
            within_bounds = False
        training_abs_bounds = self.config["training_abs_bounds"]
        for idx, pname in enumerate(
            [
                "q_mass",
                "q_light",
                "theta_E",
                "n_sersic",
                "R_sersic",
                "core_size",
                "gamma",
                "b_ani",
                "incli",
            ]
        ):
            if (
                input_p[idx] < training_abs_bounds[pname][0]
                or input_p[idx] > training_abs_bounds[pname][1]
            ):
                if verbose:
                    print(
                        "NN CALL WARNING: param",
                        pname,
                        "is outside of training bounds!",
                    )
                within_bounds = False
        if input_p[4] > input_p[2] or input_p[4] < 0.5 * input_p[2]:
            if verbose:
                print(
                    "NN CALL WARNING: R_sersic is not within training bounds- must be between 0.5R_E and R_E!"
                )
            within_bounds = False
        if (input_p[8] - 2) * np.pi / 180 < np.arccos(
            np.min([input_p[0], input_p[1]])
        ):  # wrt q_light and q_mass
            if verbose:
                print("NN CALL WARNING: Inclination is nonphysical!")
            within_bounds = False
        return within_bounds



import pytest
import numpy.testing as npt
import numpy as np

from lenstronomy.ImSim.MultiBand.single_band_multi_model import SingleBandMultiModel

from lenstronomy.Util import simulation_util
from lenstronomy.ImSim.image_model import ImageModel
from lenstronomy.LensModel.lens_model import LensModel
from lenstronomy.LightModel.light_model import LightModel
from lenstronomy.PointSource.point_source import PointSource
from lenstronomy.Data.imaging_data import ImageData
from lenstronomy.Data.psf import PSF


class TestSingleBandMultiModel(object):
    def setup_method(self):
        # data specifics
        sigma_bkg = 0.05  # background noise per pixel
        exp_time = 100  # exposure time (arbitrary units, flux per pixel is in units #photons/exp_time unit)
        numPix = 10  # cutout pixel size
        deltaPix = 0.1  # pixel size in arcsec (area per pixel = deltaPix**2)
        fwhm = 0.5  # full width half max of PSF

        # PSF specification

        kwargs_data = simulation_util.data_configure_simple(
            numPix, deltaPix, exp_time, sigma_bkg
        )
        data_class = ImageData(**kwargs_data)
        kwargs_psf = {"psf_type": "GAUSSIAN", "fwhm": fwhm, "pixel_size": deltaPix}
        psf_class = PSF(**kwargs_psf)
        kwargs_spemd = {
            "theta_E": 1.0,
            "gamma": 1.8,
            "center_x": 0,
            "center_y": 0,
            "e1": 0.1,
            "e2": 0.1,
        }

        lens_model_list = ["SPEP"]
        kwargs_lens = [kwargs_spemd]
        lens_model_class = LensModel(lens_model_list=lens_model_list)
        kwargs_sersic = {
            "amp": 1.0,
            "R_sersic": 0.1,
            "n_sersic": 2,
            "center_x": 0,
            "center_y": 0,
        }
        # 'SERSIC_ELLIPSE': elliptical Sersic profile
        kwargs_sersic_ellipse = {
            "amp": 1.0,
            "R_sersic": 0.6,
            "n_sersic": 3,
            "center_x": 0,
            "center_y": 0,
            "e1": 0.1,
            "e2": 0.1,
        }

        lens_light_model_list = ["SERSIC"]
        kwargs_lens_light = [kwargs_sersic]
        lens_light_model_class = LightModel(light_model_list=lens_light_model_list)
        source_model_list = ["SERSIC_ELLIPSE"]
        kwargs_source = [kwargs_sersic_ellipse]
        source_model_class = LightModel(light_model_list=source_model_list)

        # Point Source
        point_source_model_list = ["UNLENSED"]
        kwargs_ps = [{"ra_image": [0.4], "dec_image": [-0.2], "point_amp": [2]}]
        point_source_class = PointSource(point_source_type_list=point_source_model_list)

        kwargs_numerics = {
            "supersampling_factor": 1,
            "supersampling_convolution": False,
            "compute_mode": "regular",
        }
        imageModel = ImageModel(
            data_class,
            psf_class,
            lens_model_class,
            source_model_class,
            lens_light_model_class,
            point_source_class=point_source_class,
            kwargs_numerics=kwargs_numerics,
        )
        self.imageModel = imageModel
        image_sim = simulation_util.simulate_simple(
            imageModel, kwargs_lens, kwargs_source, kwargs_lens_light, kwargs_ps
        )

        data_class.update_data(image_sim)
        kwargs_data["image_data"] = image_sim
        self.multi_band_list = [[kwargs_data, kwargs_psf, kwargs_numerics]]

        self.kwargs_model = {
            "lens_model_list": lens_model_list,
            "source_light_model_list": source_model_list,
            "lens_light_model_list": lens_light_model_list,
            "point_source_model_list": point_source_model_list,
            "fixed_magnification_list": [False],
            "index_lens_model_list": [[0]],
            "index_lens_light_model_list": [[0]],
            "index_source_light_model_list": [[0]],
            "index_point_source_model_list": [[0]],
            "point_source_frame_list": [[0]],
        }

        self.kwargs_params = {
            "kwargs_lens": kwargs_lens,
            "kwargs_source": kwargs_source,
            "kwargs_lens_light": kwargs_lens_light,
            "kwargs_ps": kwargs_ps,
        }

        self.single_band = SingleBandMultiModel(
            multi_band_list=self.multi_band_list,
            kwargs_model=self.kwargs_model,
            linear_solver=True,
        )
        self.single_band_no_linear = SingleBandMultiModel(
            multi_band_list=self.multi_band_list,
            kwargs_model=self.kwargs_model,
            linear_solver=False,
        )

    def test_likelihood_data_given_model(self):
        logl = self.single_band.likelihood_data_given_model(**self.kwargs_params)
        logl_no_linear = self.single_band_no_linear.likelihood_data_given_model(
            **self.kwargs_params
        )
        npt.assert_almost_equal(logl / logl_no_linear, 1, decimal=4)

    def test_num_param_linear(self):
        num_linear = self.single_band.num_param_linear(**self.kwargs_params)
        assert num_linear == 3
        num_linear = self.single_band_no_linear.num_param_linear(**self.kwargs_params)
        assert num_linear == 0

    def test_image(self):
        image = self.single_band.image(**self.kwargs_params)
        image_ = self.imageModel.image(**self.kwargs_params)
        npt.assert_almost_equal(image, image_)

    def test_source_surface_brightness(self):
        image = self.single_band.source_surface_brightness(
            kwargs_source=self.kwargs_params["kwargs_source"],
            kwargs_lens=self.kwargs_params["kwargs_lens"],
        )
        image_ = self.imageModel.source_surface_brightness(
            kwargs_source=self.kwargs_params["kwargs_source"],
            kwargs_lens=self.kwargs_params["kwargs_lens"],
        )
        npt.assert_almost_equal(image, image_)

    def test_lens_surface_brightness(self):
        image = self.single_band.lens_surface_brightness(
            kwargs_lens_light=self.kwargs_params["kwargs_lens_light"]
        )
        image_ = self.imageModel.lens_surface_brightness(
            kwargs_lens_light=self.kwargs_params["kwargs_lens_light"]
        )
        npt.assert_almost_equal(image, image_)

    def test_point_source(self):
        image = self.single_band.point_source(
            kwargs_lens=self.kwargs_params["kwargs_lens"],
            kwargs_ps=self.kwargs_params["kwargs_ps"],
        )
        image_ = self.imageModel.point_source(
            kwargs_lens=self.kwargs_params["kwargs_lens"],
            kwargs_ps=self.kwargs_params["kwargs_ps"],
        )
        npt.assert_almost_equal(image, image_)

    def test_update_linear_kwargs(self):
        num = self.single_band.num_param_linear(
            self.kwargs_params["kwargs_lens"],
            self.kwargs_params["kwargs_source"],
            self.kwargs_params["kwargs_lens_light"],
            self.kwargs_params["kwargs_ps"],
        )
        param = np.ones(num) * 10
        (
            kwargs_lens,
            kwargs_source,
            kwargs_lens_light,
            kwargs_ps,
        ) = self.single_band.update_linear_kwargs(
            param,
            kwargs_lens=self.kwargs_params["kwargs_lens"],
            kwargs_source=self.kwargs_params["kwargs_ps"],
            kwargs_lens_light=self.kwargs_params["kwargs_lens_light"],
            kwargs_ps=self.kwargs_params["kwargs_ps"],
        )
        assert kwargs_source[0]["amp"] == 10

    def test_extinction_map(self):
        extinction_map = self.single_band.extinction_map(
            kwargs_extinction=None, kwargs_special=None
        )
        npt.assert_almost_equal(extinction_map, 1)


if __name__ == "__main__":
    pytest.main()



__author__ = "sibirrer"

import scipy.optimize
import scipy.interpolate as interpolate
import numpy as np

from astropy.cosmology import FlatLambdaCDM
from lenstronomy.Cosmo.lens_cosmo import LensCosmo

from lenstronomy.Util.package_util import exporter

export, __all__ = exporter()


@export
def cosmo2angular_diameter_distances(H_0, omega_m, z_lens, z_source):
    """

    :param H_0: Hubble constant [km/s/Mpc]
    :param omega_m: dimensionless matter density at z=0
    :param z_lens: deflector redshift
    :param z_source: source redshift
    :return: angular diameter distances Dd and Ds/Dds
    """
    cosmo = FlatLambdaCDM(H0=H_0, Om0=omega_m, Ob0=0.0)
    lensCosmo = LensCosmo(z_lens=z_lens, z_source=z_source, cosmo=cosmo)
    Dd = lensCosmo.dd
    Ds = lensCosmo.ds
    Dds = lensCosmo.dds
    return Dd, Ds / Dds


@export
def ddt2h0(ddt, z_lens, z_source, cosmo):
    """Converts time-delay distance to H0 for a given expansion history.

    :param ddt: time-delay distance in Mpc
    :param z_lens: deflector redshift
    :param z_source: source redshift
    :param cosmo: astropy.cosmology class instance
    :return: h0 value which matches the cosmology class effectively replacing the h0
        value used in the creation of this class
    """
    h0_fiducial = cosmo.H0.value
    lens_cosmo = LensCosmo(z_lens=z_lens, z_source=z_source, cosmo=cosmo)
    ddt_fiducial = lens_cosmo.ddt
    h0 = h0_fiducial * ddt_fiducial / ddt
    return h0


@export
class SolverFlatLCDM(object):
    """Class to solve multidimensional non-linear equations to determine the
    cosmological parameters H0 and omega_m given the angular diameter distance
    relations."""

    def __init__(self, z_d, z_s):
        self.z_d = z_d
        self.z_s = z_s

    def F(self, x, Dd, Ds_Dds):
        """

        :param x: array of parameters (H_0, omega_m)
        :return:
        """
        [H_0, omega_m] = x
        omega_m = abs(omega_m) % 1
        Dd_new, Ds_Dds_new = cosmo2angular_diameter_distances(
            H_0, omega_m, self.z_d, self.z_s
        )
        y = np.zeros(2)
        y[0] = Dd - Dd_new
        y[1] = Ds_Dds - Ds_Dds_new
        return y

    def solve(self, init, dd, ds_dds):
        x = scipy.optimize.fsolve(
            self.F, init, args=(dd, ds_dds), xtol=1.49012e-08, factor=0.1
        )
        x[1] = abs(x[1]) % 1
        y = self.F(x, dd, ds_dds)
        if abs(y[0]) >= 0.1 or abs(y[1]) > 0.1:
            x = np.array([-1, -1])
        return x


@export
class InvertCosmo(object):
    """Class to do an interpolation and call the inverse of this interpolation to get
    H_0 and omega_m."""

    def __init__(self, z_d, z_s, H0_range=None, omega_m_range=None):
        self.z_d = z_d
        self.z_s = z_s
        if H0_range is None:
            H0_range = np.linspace(10, 100, 90)
        if omega_m_range is None:
            omega_m_range = np.linspace(0.05, 1, 95)
        self._H0_range = H0_range
        self._omega_m_range = omega_m_range

    def _make_interpolation(self):
        """Creates an interpolation grid in H_0, omega_m and computes quantities in Dd
        and Ds_Dds.

        :return:
        """
        grid2d = np.dstack(np.meshgrid(self._H0_range, self._omega_m_range)).reshape(
            -1, 2
        )
        H0_grid = grid2d[:, 0]
        omega_m_grid = grid2d[:, 1]
        Dd_grid = np.zeros_like(H0_grid)
        Ds_Dds_grid = np.zeros_like(H0_grid)
        for i in range(len(H0_grid)):
            Dd, Ds_Dds = cosmo2angular_diameter_distances(
                H0_grid[i], omega_m_grid[i], self.z_d, self.z_s
            )
            Dd_grid[i] = Dd
            Ds_Dds_grid[i] = Ds_Dds
        self._f_H0 = interpolate.interp2d(
            Dd_grid,
            Ds_Dds_grid,
            H0_grid,
            kind="linear",
            copy=False,
            bounds_error=False,
            fill_value=-1,
        )
        print("H0 interpolation done")
        self._f_omega_m = interpolate.interp2d(
            Dd_grid,
            Ds_Dds_grid,
            omega_m_grid,
            kind="linear",
            copy=False,
            bounds_error=False,
            fill_value=0,
        )
        print("omega_m interpolation done")

    def get_cosmo(self, Dd, Ds_Dds):
        """Return the values of H0 and omega_m computed with an interpolation.

        :param Dd: flat
        :param Ds_Dds: float
        :return:
        """
        if not hasattr(self, "_f_H0") or not hasattr(self, "_f_omega_m"):
            self._make_interpolation()
        H0 = self._f_H0(Dd, Ds_Dds)
        print(H0, "H0")
        omega_m = self._f_omega_m(Dd, Ds_Dds)
        Dd_new, Ds_Dds_new = cosmo2angular_diameter_distances(
            H0[0], omega_m[0], self.z_d, self.z_s
        )
        if abs(Dd - Dd_new) / Dd > 0.01 or abs(Ds_Dds - Ds_Dds_new) / Ds_Dds > 0.01:
            return -1, -1
        else:
            return H0[0], omega_m[0]



__author__ = "sibirrer"

from scipy import fftpack
import numpy as np
import lenstronomy.Util.analysis_util as analysis_util

from lenstronomy.Util.package_util import exporter

export, __all__ = exporter()


@export
def correlation_2D(image):
    """#TODO document normalization output in units

    :param image: 2d image
    :return: 2d fourier transform
    """
    # Take the fourier transform of the image.
    f1 = fftpack.fft2(image)

    # Now shift the quadrants around so that low spatial frequencies are in
    # the center of the 2D fourier transformed image.
    f2 = fftpack.fftshift(f1)
    return np.abs(f2)


@export
def power_spectrum_2d(image):
    """

    :param image: 2d numpy array
    :return: 2d power spectrum in frequency units of the pixels
    """
    nx, ny = np.shape(image)
    corr2d = correlation_2D(image)
    return (corr2d / nx / ny) ** 2


@export
def power_spectrum_1d(image):
    """

    :param image: 2d numpy array
    :return: 1d radially averaged power spectrum of image in frequency units of pixels, radius in units of pixels
    """
    psd2D = power_spectrum_2d(image)
    # Calculate the azimuthally averaged 1D power spectrum
    psd1D, r = analysis_util.azimuthalAverage(psd2D)
    return psd1D, r



from lenstronomy.LightModel.Profiles.shapelets import ShapeletSet
from lenstronomy.Util import param_util


class ShapeletSetEllipse(object):
    """Cartesian shapelets with elliptical axis ratios."""

    param_names = ["amp", "n_max", "beta", "e1", "e2", "center_x", "center_y"]
    lower_limit_default = {
        "beta": 0.01,
        "e1": -0.6,
        "e2": -0.6,
        "center_x": -100,
        "center_y": -100,
    }
    upper_limit_default = {
        "beta": 100,
        "e1": 0.6,
        "e2": 0.6,
        "center_x": 100,
        "center_y": 100,
    }

    def __init__(self):
        self._shapelet_set = ShapeletSet()

    def function(self, x, y, amp, n_max, beta, e1, e2, center_x=0, center_y=0):
        """

        :param x: x-coordinates
        :param y: y-coordinates
        :param amp: array of amplitudes in pre-defined order of shapelet basis functions
        :param beta: shapelet scale
        :param n_max: maximum polynomial order in Hermite polynomial
        :param e1: eccentricity component 1
        :param e2: eccentricity component 2
        :param center_x: shapelet center x
        :param center_y: shapelet center y
        :return: surface brightness of combined shapelet set
        """
        x_, y_ = param_util.transform_e1e2_product_average(
            x, y, e1, e2, center_x=0, center_y=0
        )
        return self._shapelet_set.function(x_, y_, amp, n_max, beta, center_x, center_y)

    def function_split(self, x, y, amp, n_max, beta, e1, e2, center_x=0, center_y=0):
        """Splits shapelet set in list of individual shapelet basis function responses.

        :param x: x-coordinates
        :param y: y-coordinates
        :param amp: array of amplitudes in pre-defined order of shapelet basis functions
        :param beta: shapelet scale
        :param n_max: maximum polynomial order in Hermite polynomial
        :param e1: eccentricity component 1
        :param e2: eccentricity component 2
        :param center_x: shapelet center x
        :param center_y: shapelet center y
        :return: list of individual shapelet basis function responses
        """
        x_, y_ = param_util.transform_e1e2_product_average(
            x, y, e1, e2, center_x=0, center_y=0
        )
        return self._shapelet_set.function_split(
            x_, y_, amp, n_max, beta, center_x, center_y
        )



__author__ = "sibirrer"

import numpy.testing as npt
import numpy as np
import pytest
import unittest

from lenstronomy.Analysis.kinematics_api import KinematicsAPI
import lenstronomy.Util.param_util as param_util


class TestKinematicsAPI(object):
    def setup_method(self):
        pass

    def test_velocity_dispersion(self):
        z_lens = 0.5
        z_source = 1.5
        kwargs_model = {
            "lens_model_list": ["SPEP", "SHEAR", "SIS", "SIS", "SIS"],
            "lens_light_model_list": ["SERSIC_ELLIPSE", "SERSIC"],
        }

        theta_E = 1.5
        gamma = 1.8
        kwargs_lens = [
            {
                "theta_E": theta_E,
                "e1": 0,
                "center_x": -0.044798916793300093,
                "center_y": 0.0054408937891703788,
                "e2": 0,
                "gamma": gamma,
            },
            {"e1": -0.050871696555354479, "e2": -0.0061601733920590464},
            {
                "center_y": 2.79985456,
                "center_x": -2.32019894,
                "theta_E": 0.28165274714097904,
            },
            {
                "center_y": 3.83985426,
                "center_x": -2.32019933,
                "theta_E": 0.0038110812674654873,
            },
            {
                "center_y": 4.31985428,
                "center_x": -1.68019931,
                "theta_E": 0.45552039839735037,
            },
        ]

        phi, q = -0.52624727893702705, 0.79703498156919605
        e1, e2 = param_util.phi_q2_ellipticity(phi, q)
        kwargs_lens_light = [
            {
                "n_sersic": 1.1212528655709217,
                "center_x": -0.019674496231393473,
                "e1": e1,
                "e2": e2,
                "amp": 1.1091367792010356,
                "center_y": 0.076914975081560991,
                "R_sersic": 0.42691611878867058,
            },
            {
                "R_sersic": 0.03025682660635394,
                "amp": 139.96763298885992,
                "n_sersic": 1.90000008624093865,
                "center_x": -0.019674496231393473,
                "center_y": 0.076914975081560991,
            },
        ]
        r_ani = 0.62
        kwargs_anisotropy = {"r_ani": r_ani}
        R_slit = 3.8
        dR_slit = 1.0
        aperture_type = "slit"
        kwargs_aperture = {
            "aperture_type": aperture_type,
            "center_ra": 0,
            "width": dR_slit,
            "length": R_slit,
            "angle": 0,
            "center_dec": 0,
        }

        psf_fwhm = 0.7
        kwargs_psf = {"psf_type": "GAUSSIAN", "fwhm": psf_fwhm}
        anisotropy_model = "OM"
        kwargs_mge = {"n_comp": 20}
        r_eff = 0.211919902322
        kinematicAPI = KinematicsAPI(
            z_lens,
            z_source,
            kwargs_model,
            kwargs_aperture=kwargs_aperture,
            kwargs_seeing=kwargs_psf,
            lens_model_kinematics_bool=[True, False, False, False, False],
            anisotropy_model=anisotropy_model,
            kwargs_mge_light=kwargs_mge,
            kwargs_mge_mass=kwargs_mge,
            sampling_number=1000,
            MGE_light=True,
        )

        v_sigma = kinematicAPI.velocity_dispersion(
            kwargs_lens, kwargs_lens_light, kwargs_anisotropy, r_eff=r_eff
        )

        kinematicAPI = KinematicsAPI(
            z_lens,
            z_source,
            kwargs_model,
            kwargs_aperture=kwargs_aperture,
            kwargs_seeing=kwargs_psf,
            lens_model_kinematics_bool=[True, False, False, False, False],
            anisotropy_model=anisotropy_model,
            kwargs_mge_light=kwargs_mge,
            kwargs_mge_mass=kwargs_mge,
            sampling_number=1000,
            MGE_light=True,
            MGE_mass=True,
        )
        v_sigma_mge_lens = kinematicAPI.velocity_dispersion(
            kwargs_lens,
            kwargs_lens_light,
            kwargs_anisotropy,
            r_eff=r_eff,
            theta_E=theta_E,
        )
        # v_sigma_mge_lens = kinematicAPI.velocity_dispersion_numerical(kwargs_lens, kwargs_lens_light, kwargs_anisotropy, kwargs_aperture,
        #                                                          kwargs_psf, anisotropy_model, MGE_light=True, MGE_mass=True, theta_E=theta_E,
        #                                                          kwargs_mge_light=kwargs_mge, kwargs_mge_mass=kwargs_mge,
        #                                                          r_eff=r_eff)
        kinematicAPI = KinematicsAPI(
            z_lens,
            z_source,
            kwargs_model,
            kwargs_aperture=kwargs_aperture,
            kwargs_seeing=kwargs_psf,
            lens_model_kinematics_bool=[True, False, False, False, False],
            anisotropy_model=anisotropy_model,
            kwargs_mge_light=kwargs_mge,
            kwargs_mge_mass=kwargs_mge,
            sampling_number=1000,
            MGE_light=False,
            MGE_mass=False,
            Hernquist_approx=True,
        )
        v_sigma_hernquist = kinematicAPI.velocity_dispersion(
            kwargs_lens,
            kwargs_lens_light,
            kwargs_anisotropy,
            r_eff=r_eff,
            theta_E=theta_E,
        )
        # v_sigma_hernquist = kinematicAPI.velocity_dispersion_numerical(kwargs_lens, kwargs_lens_light, kwargs_anisotropy,
        #                                                          kwargs_aperture, kwargs_psf, anisotropy_model,
        #                                                          MGE_light=False, MGE_mass=False,
        #                                                          r_eff=r_eff, Hernquist_approx=True)

        vel_disp_temp = kinematicAPI.velocity_dispersion_analytical(
            theta_E, gamma, r_ani=r_ani, r_eff=r_eff
        )
        print(v_sigma, vel_disp_temp)
        # assert 1 == 0
        npt.assert_almost_equal(v_sigma / vel_disp_temp, 1, decimal=1)
        npt.assert_almost_equal(v_sigma_mge_lens / v_sigma, 1, decimal=1)
        npt.assert_almost_equal(v_sigma / v_sigma_hernquist, 1, decimal=1)

    def test_galkin_settings(self):
        z_lens = 0.5
        z_source = 1.5
        kwargs_model = {
            "lens_model_list": ["SIS"],
            "lens_light_model_list": ["HERNQUIST"],
        }

        kwargs_lens = [{"theta_E": 1, "center_x": 0, "center_y": 0}]
        kwargs_lens_light = [{"amp": 1, "Rs": 1, "center_x": 0, "center_y": 0}]
        r_ani = 0.62
        kwargs_anisotropy = {"r_ani": r_ani}
        R_slit = 3.8
        dR_slit = 1.0
        aperture_type = "slit"
        kwargs_aperture = {
            "aperture_type": aperture_type,
            "center_ra": 0,
            "width": dR_slit,
            "length": R_slit,
            "angle": 0,
            "center_dec": 0,
        }

        psf_fwhm = 0.7
        kwargs_psf = {"psf_type": "GAUSSIAN", "fwhm": psf_fwhm}
        anisotropy_model = "OM"
        kwargs_mge = {"n_comp": 20}
        kinematicAPI = KinematicsAPI(
            z_lens,
            z_source,
            kwargs_model,
            kwargs_aperture=kwargs_aperture,
            kwargs_seeing=kwargs_psf,
            analytic_kinematics=True,
            anisotropy_model=anisotropy_model,
            kwargs_mge_light=kwargs_mge,
            kwargs_mge_mass=kwargs_mge,
            sampling_number=1000,
        )
        galkin, kwargs_profile, kwargs_light = kinematicAPI.galkin_settings(
            kwargs_lens, kwargs_lens_light, r_eff=None, theta_E=None, gamma=None
        )
        npt.assert_almost_equal(kwargs_profile["gamma"], 2, decimal=2)

        kinematicAPI = KinematicsAPI(
            z_lens,
            z_source,
            kwargs_model,
            kwargs_aperture=[kwargs_aperture],
            kwargs_seeing=[kwargs_psf],
            analytic_kinematics=True,
            anisotropy_model=anisotropy_model,
            multi_observations=True,
            kwargs_mge_light=kwargs_mge,
            kwargs_mge_mass=kwargs_mge,
            sampling_number=1000,
        )
        galkin, kwargs_profile, kwargs_light = kinematicAPI.galkin_settings(
            kwargs_lens, kwargs_lens_light, r_eff=None, theta_E=None, gamma=None
        )
        npt.assert_almost_equal(kwargs_profile["gamma"], 2, decimal=2)

    def test_kinematic_light_profile(self):
        z_lens = 0.5
        z_source = 1.5
        kwargs_options = {"lens_light_model_list": ["HERNQUIST_ELLIPSE", "SERSIC"]}
        kwargs_mge = {"n_comp": 20}
        kinematicAPI = KinematicsAPI(
            z_lens,
            z_source,
            kwargs_options,
            kwargs_seeing={},
            kwargs_aperture={},
            anisotropy_model="OM",
        )
        r_eff = 0.2
        kwargs_lens_light = [
            {
                "amp": 1,
                "Rs": r_eff * 0.551,
                "e1": 0.0,
                "e2": 0,
                "center_x": 0,
                "center_y": 0,
            },
            {"amp": 1, "R_sersic": 1, "n_sersic": 2, "center_x": -10, "center_y": -10},
        ]
        light_profile_list, kwargs_light = kinematicAPI.kinematic_light_profile(
            kwargs_lens_light,
            MGE_fit=True,
            r_eff=r_eff,
            model_kinematics_bool=[True, False],
            kwargs_mge=kwargs_mge,
        )
        assert light_profile_list[0] == "MULTI_GAUSSIAN"

        light_profile_list, kwargs_light = kinematicAPI.kinematic_light_profile(
            kwargs_lens_light,
            MGE_fit=False,
            r_eff=r_eff,
            model_kinematics_bool=[True, False],
        )
        assert light_profile_list[0] == "HERNQUIST_ELLIPSE"

        light_profile_list, kwargs_light = kinematicAPI.kinematic_light_profile(
            kwargs_lens_light,
            MGE_fit=False,
            Hernquist_approx=True,
            r_eff=r_eff,
            model_kinematics_bool=[True, False],
        )
        assert light_profile_list[0] == "HERNQUIST"
        npt.assert_almost_equal(
            kwargs_light[0]["Rs"] / kwargs_lens_light[0]["Rs"], 1, decimal=2
        )

    def test_kinematic_lens_profiles(self):
        z_lens = 0.5
        z_source = 1.5
        kwargs_options = {"lens_model_list": ["SPEP", "SHEAR"]}
        kin_api = KinematicsAPI(
            z_lens,
            z_source,
            kwargs_options,
            kwargs_aperture={},
            kwargs_seeing={},
            anisotropy_model="OM",
        )
        kwargs_lens = [
            {
                "theta_E": 1.4272358196260446,
                "e1": 0,
                "center_x": -0.044798916793300093,
                "center_y": 0.0054408937891703788,
                "e2": 0,
                "gamma": 1.8,
            },
            {"e1": -0.050871696555354479, "e2": -0.0061601733920590464},
        ]

        kwargs_mge = {"n_comp": 20}
        mass_profile_list, kwargs_profile = kin_api.kinematic_lens_profiles(
            kwargs_lens,
            MGE_fit=True,
            kwargs_mge=kwargs_mge,
            theta_E=1.4,
            model_kinematics_bool=[True, False],
        )
        assert mass_profile_list[0] == "MULTI_GAUSSIAN_KAPPA"

        mass_profile_list, kwargs_profile = kin_api.kinematic_lens_profiles(
            kwargs_lens, MGE_fit=False, model_kinematics_bool=[True, False]
        )
        assert mass_profile_list[0] == "SPEP"

    def test_model_dispersion(self):
        np.random.seed(42)
        z_lens = 0.5
        z_source = 1.5
        r_eff = 1.0
        theta_E = 1.0
        kwargs_model = {
            "lens_model_list": ["SIS"],
            "lens_light_model_list": ["HERNQUIST"],
        }
        kwargs_lens = [{"theta_E": theta_E, "center_x": 0, "center_y": 0}]
        kwargs_lens_light = [
            {"amp": 1, "Rs": r_eff * 0.551, "center_x": 0, "center_y": 0}
        ]
        kwargs_anisotropy = {"r_ani": 1}
        # settings

        R_slit = 3.8
        dR_slit = 1.0
        aperture_type = "slit"
        kwargs_aperture = {
            "aperture_type": aperture_type,
            "center_ra": 0,
            "width": dR_slit,
            "length": R_slit,
            "angle": 0,
            "center_dec": 0,
        }
        psf_fwhm = 0.7
        kwargs_seeing = {"psf_type": "GAUSSIAN", "fwhm": psf_fwhm}
        anisotropy_model = "OM"
        kin_api = KinematicsAPI(
            z_lens,
            z_source,
            kwargs_model,
            kwargs_aperture,
            kwargs_seeing,
            anisotropy_model=anisotropy_model,
        )

        kwargs_numerics_galkin = {
            "interpol_grid_num": 2000,
            "log_integration": True,
            "max_integrate": 1000,
            "min_integrate": 0.0001,
        }
        kin_api.kinematics_modeling_settings(
            anisotropy_model,
            kwargs_numerics_galkin,
            analytic_kinematics=True,
            Hernquist_approx=False,
            MGE_light=False,
            MGE_mass=False,
        )
        vel_disp_analytic = kin_api.velocity_dispersion(
            kwargs_lens,
            kwargs_lens_light,
            kwargs_anisotropy,
            r_eff=r_eff,
            theta_E=theta_E,
            gamma=2,
        )

        kin_api.kinematics_modeling_settings(
            anisotropy_model,
            kwargs_numerics_galkin,
            analytic_kinematics=False,
            Hernquist_approx=False,
            MGE_light=False,
            MGE_mass=False,
        )
        vel_disp_numerical = kin_api.velocity_dispersion(
            kwargs_lens, kwargs_lens_light, kwargs_anisotropy
        )  # ,
        # r_eff=r_eff, theta_E=theta_E, gamma=2)
        npt.assert_almost_equal(vel_disp_numerical / vel_disp_analytic, 1, decimal=2)

        kin_api.kinematics_modeling_settings(
            anisotropy_model,
            kwargs_numerics_galkin,
            analytic_kinematics=False,
            Hernquist_approx=False,
            MGE_light=False,
            MGE_mass=False,
            kwargs_mge_light={"n_comp": 10},
            kwargs_mge_mass={"n_comp": 5},
        )
        assert kin_api._kwargs_mge_mass["n_comp"] == 5
        assert kin_api._kwargs_mge_light["n_comp"] == 10

    def test_velocity_dispersion_map(self):
        np.random.seed(42)
        z_lens = 0.5
        z_source = 1.5
        kwargs_options = {
            "lens_model_list": ["SIS"],
            "lens_light_model_list": ["HERNQUIST"],
        }
        r_eff = 1.0
        theta_E = 1
        kwargs_lens = [{"theta_E": theta_E, "center_x": 0, "center_y": 0}]
        kwargs_lens_light = [
            {"amp": 1, "Rs": r_eff * 0.551, "center_x": 0, "center_y": 0}
        ]
        kwargs_anisotropy = {"r_ani": 1}

        r_bins = np.array([0, 0.5, 1])
        aperture_type = "IFU_shells"
        kwargs_aperture = {
            "aperture_type": aperture_type,
            "center_ra": 0,
            "r_bins": r_bins,
            "center_dec": 0,
        }
        psf_fwhm = 0.7
        kwargs_seeing = {"psf_type": "GAUSSIAN", "fwhm": psf_fwhm}
        anisotropy_model = "OM"
        kin_api = KinematicsAPI(
            z_lens,
            z_source,
            kwargs_options,
            kwargs_aperture=kwargs_aperture,
            kwargs_seeing=kwargs_seeing,
            anisotropy_model=anisotropy_model,
        )

        kwargs_numerics_galkin = {
            "interpol_grid_num": 500,
            "log_integration": True,
            "max_integrate": 10,
            "min_integrate": 0.001,
        }
        kin_api.kinematics_modeling_settings(
            anisotropy_model,
            kwargs_numerics_galkin,
            analytic_kinematics=True,
            Hernquist_approx=False,
            MGE_light=False,
            MGE_mass=False,
            num_kin_sampling=1000,
            num_psf_sampling=100,
        )
        vel_disp_analytic = kin_api.velocity_dispersion_map(
            kwargs_lens,
            kwargs_lens_light,
            kwargs_anisotropy,
            r_eff=r_eff,
            theta_E=theta_E,
            gamma=2,
        )

        kin_api.kinematics_modeling_settings(
            anisotropy_model,
            kwargs_numerics_galkin,
            analytic_kinematics=False,
            Hernquist_approx=False,
            MGE_light=False,
            MGE_mass=False,
            num_kin_sampling=1000,
            num_psf_sampling=100,
        )
        vel_disp_numerical = kin_api.velocity_dispersion_map(
            kwargs_lens,
            kwargs_lens_light,
            kwargs_anisotropy,
            r_eff=r_eff,
            theta_E=theta_E,
            gamma=2,
        )
        print(vel_disp_numerical, vel_disp_analytic)
        npt.assert_almost_equal(vel_disp_numerical, vel_disp_analytic, decimal=-1)

    def test_interpolated_sersic(self):
        from lenstronomy.Analysis.light2mass import light2mass_interpol

        kwargs_light = [
            {
                "n_sersic": 2,
                "R_sersic": 0.5,
                "amp": 1,
                "center_x": 0.01,
                "center_y": 0.01,
            }
        ]
        kwargs_lens = [
            {
                "n_sersic": 2,
                "R_sersic": 0.5,
                "k_eff": 1,
                "center_x": 0.01,
                "center_y": 0.01,
            }
        ]
        deltaPix = 0.1
        numPix = 100

        kwargs_interp = light2mass_interpol(
            ["SERSIC"],
            kwargs_lens_light=kwargs_light,
            numPix=numPix,
            deltaPix=deltaPix,
            subgrid_res=5,
        )
        kwargs_lens_interp = [kwargs_interp]
        from lenstronomy.Analysis.kinematics_api import KinematicsAPI

        z_lens = 0.5
        z_source = 1.5
        r_ani = 0.62
        kwargs_anisotropy = {"r_ani": r_ani}
        R_slit = 3.8
        dR_slit = 1.0
        aperture_type = "slit"
        kwargs_aperture = {
            "center_ra": 0,
            "width": dR_slit,
            "length": R_slit,
            "angle": 0,
            "center_dec": 0,
            "aperture_type": aperture_type,
        }
        psf_fwhm = 0.7
        kwargs_psf = {"psf_type": "GAUSSIAN", "fwhm": psf_fwhm}
        anisotropy_model = "OM"
        r_eff = 0.5
        kwargs_model = {
            "lens_model_list": ["SERSIC"],
            "lens_light_model_list": ["SERSIC"],
        }
        kwargs_mge = {"n_comp": 20}
        kinematic_api = KinematicsAPI(
            z_lens,
            z_source,
            kwargs_model,
            kwargs_aperture,
            kwargs_seeing=kwargs_psf,
            anisotropy_model=anisotropy_model,
            MGE_light=True,
            MGE_mass=True,
            kwargs_mge_mass=kwargs_mge,
            kwargs_mge_light=kwargs_mge,
        )

        v_sigma = kinematic_api.velocity_dispersion(
            kwargs_lens, kwargs_light, kwargs_anisotropy, r_eff=r_eff, theta_E=1
        )
        kwargs_model_interp = {
            "lens_model_list": ["INTERPOL"],
            "lens_light_model_list": ["SERSIC"],
        }
        kinematic_api_interp = KinematicsAPI(
            z_lens,
            z_source,
            kwargs_model_interp,
            kwargs_aperture,
            kwargs_seeing=kwargs_psf,
            anisotropy_model=anisotropy_model,
            MGE_light=True,
            MGE_mass=True,
            kwargs_mge_mass=kwargs_mge,
            kwargs_mge_light=kwargs_mge,
        )
        v_sigma_interp = kinematic_api_interp.velocity_dispersion(
            kwargs_lens_interp,
            kwargs_light,
            kwargs_anisotropy,
            theta_E=1.0,
            r_eff=r_eff,
        )
        npt.assert_almost_equal(v_sigma / v_sigma_interp, 1, 1)
        # use as kinematic constraints
        # compare with MGE Sersic kinematic estimate


class TestRaise(unittest.TestCase):
    def test_raise(self):
        with self.assertRaises(ValueError):
            z_lens = 0.5
            z_source = 1.5
            kwargs_model = {"lens_light_model_list": ["HERNQUIST"]}
            kinematicAPI = KinematicsAPI(
                z_lens,
                z_source,
                kwargs_model,
                kwargs_seeing={},
                kwargs_aperture={},
                anisotropy_model="OM",
            )
            kwargs_light = [{"Rs": 1, "amp": 1, "center_x": 0, "center_y": 0}]
            kinematicAPI.kinematic_light_profile(
                kwargs_light,
                MGE_fit=False,
                Hernquist_approx=True,
                r_eff=None,
                model_kinematics_bool=[True],
            )
        with self.assertRaises(ValueError):
            z_lens = 0.5
            z_source = 1.5
            kwargs_model = {"lens_light_model_list": ["HERNQUIST"]}
            kinematicAPI = KinematicsAPI(
                z_lens,
                z_source,
                kwargs_model,
                kwargs_seeing={},
                kwargs_aperture={},
                anisotropy_model="OM",
            )
            kwargs_light = [{"Rs": 1, "amp": 1, "center_x": 0, "center_y": 0}]
            kinematicAPI.kinematic_light_profile(
                kwargs_light,
                MGE_fit=False,
                Hernquist_approx=False,
                r_eff=None,
                analytic_kinematics=True,
            )
        with self.assertRaises(ValueError):
            z_lens = 0.5
            z_source = 1.5
            kwargs_model = {
                "lens_light_model_list": ["HERNQUIST"],
                "lens_model_list": [],
            }
            kinematicAPI = KinematicsAPI(
                z_lens,
                z_source,
                kwargs_model,
                kwargs_seeing={},
                kwargs_aperture={},
                anisotropy_model="OM",
            )
            kwargs_light = [{"Rs": 1, "amp": 1, "center_x": 0, "center_y": 0}]
            kinematicAPI.kinematic_lens_profiles(
                kwargs_light, MGE_fit=True, model_kinematics_bool=[True]
            )
        with self.assertRaises(ValueError):
            z_lens = 0.5
            z_source = 1.5
            kwargs_model = {
                "lens_light_model_list": ["HERNQUIST"],
                "lens_model_list": [],
            }
            kinematicAPI = KinematicsAPI(
                z_lens,
                z_source,
                kwargs_model,
                kwargs_seeing={},
                kwargs_aperture={},
                anisotropy_model="OM",
            )
            kinematicAPI.kinematic_lens_profiles(
                kwargs_lens=None, analytic_kinematics=True
            )

        with self.assertRaises(ValueError):
            z_lens = 0.5
            z_source = 1.5
            kwargs_model = {
                "lens_light_model_list": ["HERNQUIST"],
                "lens_model_list": [],
            }
            kinematicAPI = KinematicsAPI(
                z_lens,
                z_source,
                kwargs_model,
                kwargs_seeing={},
                kwargs_aperture={},
                anisotropy_model="OM",
            )
            kwargs_lens_light = [{"Rs": 1, "center_x": 0, "center_y": 0}]
            kinematicAPI.kinematic_light_profile(
                kwargs_lens_light,
                r_eff=None,
                MGE_fit=True,
                model_kinematics_bool=None,
                Hernquist_approx=False,
                kwargs_mge=None,
            )
        with self.assertRaises(ValueError):
            z_lens = 0.5
            z_source = 1.5
            kwargs_model = {
                "lens_light_model_list": ["HERNQUIST"],
                "lens_model_list": ["SIS"],
            }
            kwargs_lens = [{"theta_E": 1, "center_x": 0, "center_y": 0}]
            kinematicAPI = KinematicsAPI(
                z_lens,
                z_source,
                kwargs_model,
                kwargs_seeing={},
                kwargs_aperture={},
                anisotropy_model="OM",
            )
            kinematicAPI.kinematic_lens_profiles(
                kwargs_lens,
                MGE_fit=True,
                model_kinematics_bool=None,
                theta_E=None,
                kwargs_mge={},
            )


if __name__ == "__main__":
    pytest.main()



__author__ = "sibirrer"


import numpy as np
import lenstronomy.Util.param_util as param_util
from lenstronomy.Util import util
from lenstronomy.LensModel.Profiles.base_profile import LensProfileBase
from lenstronomy.LensModel.Profiles.spp import SPP

__all__ = ["SPEP"]


class SPEP(LensProfileBase):
    """Class for Softened power-law elliptical potential (SPEP)"""

    param_names = ["theta_E", "gamma", "e1", "e2", "center_x", "center_y"]
    lower_limit_default = {
        "theta_E": 0,
        "gamma": 0,
        "e1": -0.5,
        "e2": -0.5,
        "center_x": -100,
        "center_y": -100,
    }
    upper_limit_default = {
        "theta_E": 100,
        "gamma": 100,
        "e1": 0.5,
        "e2": 0.5,
        "center_x": 100,
        "center_y": 100,
    }

    def __init__(self):
        self.spp = SPP()
        super(SPEP, self).__init__()

    def function(self, x, y, theta_E, gamma, e1, e2, center_x=0, center_y=0):
        """
        :param x: set of x-coordinates
        :type x: array of size (n)
        :param theta_E: Einstein radius of lense
        :type theta_E: float.
        :param gamma: power law slope of mass profifle
        :type gamma: <2 float
        :param e1: eccentricity
        :type e1: -1<e1<1
        :param e2: eccentricity
        :type e2: -1<e1<1
        :returns:  function
        :raises: AttributeError, KeyError
        """
        phi_G, q = param_util.ellipticity2phi_q(e1, e2)
        gamma, q = self._param_bounds(gamma, q)
        theta_E *= q
        x_shift = x - center_x
        y_shift = y - center_y
        E = theta_E / (((3 - gamma) / 2.0) ** (1.0 / (1 - gamma)) * np.sqrt(q))
        # E = phi_E
        eta = -gamma + 3
        xt1 = np.cos(phi_G) * x_shift + np.sin(phi_G) * y_shift
        xt2 = -np.sin(phi_G) * x_shift + np.cos(phi_G) * y_shift
        p2 = xt1**2 + xt2**2 / q**2
        s2 = 0.0  # softening
        return 2 * E**2 / eta**2 * ((p2 + s2) / E**2) ** (eta / 2)

    def derivatives(self, x, y, theta_E, gamma, e1, e2, center_x=0, center_y=0):
        phi_G, q = param_util.ellipticity2phi_q(e1, e2)
        gamma, q = self._param_bounds(gamma, q)
        phi_E_new = theta_E * q
        x_shift = x - center_x
        y_shift = y - center_y
        E = phi_E_new / (((3 - gamma) / 2.0) ** (1.0 / (1 - gamma)) * np.sqrt(q))
        # E = phi_E
        eta = float(-gamma + 3)
        cos_phi = np.cos(phi_G)
        sin_phi = np.sin(phi_G)

        xt1 = cos_phi * x_shift + sin_phi * y_shift
        xt2 = -sin_phi * x_shift + cos_phi * y_shift
        xt2difq2 = xt2 / (q * q)
        P2 = xt1 * xt1 + xt2 * xt2difq2
        if isinstance(P2, int) or isinstance(P2, float):
            a = max(0.000001, P2)
        else:
            a = np.empty_like(P2)
            p2 = P2[P2 > 0]  # in the SIS regime
            a[P2 == 0] = 0.000001
            a[P2 > 0] = p2
        fac = 1.0 / eta * (a / (E * E)) ** (eta / 2 - 1) * 2
        f_x_prim = fac * xt1
        f_y_prim = fac * xt2difq2

        f_x = cos_phi * f_x_prim - sin_phi * f_y_prim
        f_y = sin_phi * f_x_prim + cos_phi * f_y_prim
        return f_x, f_y

    def hessian(self, x, y, theta_E, gamma, e1, e2, center_x=0, center_y=0):
        phi_G, q = param_util.ellipticity2phi_q(e1, e2)
        gamma, q = self._param_bounds(gamma, q)
        phi_E_new = theta_E * q
        # x_shift = x - center_x
        # y_shift = y - center_y

        # shift
        x_ = x - center_x
        y_ = y - center_y
        # rotate
        x__, y__ = util.rotate(x_, y_, phi_G)

        E = phi_E_new / (((3 - gamma) / 2.0) ** (1.0 / (1 - gamma)) * np.sqrt(q))
        if E <= 0:
            return (
                np.zeros_like(x),
                np.zeros_like(x),
                np.zeros_like(x),
                np.zeros_like(x),
            )
        # E = phi_E
        eta = float(-gamma + 3)
        # xt1 = np.cos(phi_G)*x_shift+np.sin(phi_G)*y_shift
        # xt2 = -np.sin(phi_G)*x_shift+np.cos(phi_G)*y_shift
        xt1, xt2 = x__, y__
        P2 = xt1**2 + xt2**2 / q**2

        if isinstance(P2, int) or isinstance(P2, float):
            a = max(0.000001, P2)
        else:
            a = np.empty_like(P2)
            p2 = P2[P2 > 0]  # in the SIS regime
            a[P2 == 0] = 0.000001
            a[P2 > 0] = p2
        s2 = 0.0  # softening

        kappa = (
            1.0
            / eta
            * (a / E**2) ** (eta / 2 - 1)
            * ((eta - 2) * (xt1**2 + xt2**2 / q**4) / a + (1 + 1 / q**2))
        )
        gamma1_value = (
            1.0
            / eta
            * (a / E**2) ** (eta / 2 - 1)
            * (
                1
                - 1 / q**2
                + (eta / 2 - 1) * (2 * xt1**2 - 2 * xt2**2 / q**4) / a
            )
        )
        gamma2_value = (
            4
            * xt1
            * xt2
            / q**2
            * (1.0 / 2 - 1 / eta)
            * (a / E**2) ** (eta / 2 - 2)
            / E**2
        )

        gamma1 = np.cos(2 * phi_G) * gamma1_value - np.sin(2 * phi_G) * gamma2_value
        gamma2 = +np.sin(2 * phi_G) * gamma1_value + np.cos(2 * phi_G) * gamma2_value

        f_xx = kappa + gamma1
        f_yy = kappa - gamma1
        f_xy = gamma2
        return f_xx, f_xy, f_xy, f_yy

    def mass_3d_lens(self, r, theta_E, gamma, e1=None, e2=None):
        """Computes the spherical power-law mass enclosed (with SPP routine)

        :param r: radius within the mass is computed
        :param theta_E: Einstein radius
        :param gamma: power-law slope
        :param e1: eccentricity component (not used)
        :param e2: eccentricity component (not used)
        :return: mass enclosed a 3D radius r
        """
        return self.spp.mass_3d_lens(r, theta_E, gamma)

    def density_lens(self, r, theta_E, gamma, e1=None, e2=None):
        """Computes the density at 3d radius r given lens model parameterization. The
        integral in the LOS projection of this quantity results in the convergence
        quantity.

        :param r: radius within the mass is computed
        :param theta_E: Einstein radius
        :param gamma: power-law slope
        :param e1: eccentricity component (not used)
        :param e2: eccentricity component (not used)
        :return: mass enclosed a 3D radius r
        """
        return self.spp.density_lens(r, theta_E, gamma)

    @staticmethod
    def _param_bounds(gamma, q):
        """Bounds parameters.

        :param gamma:
        :param q:
        :return:
        """
        if gamma < 1.4:
            gamma = 1.4
        if gamma > 2.9:
            gamma = 2.9
        if q < 0.01:
            q = 0.01
        return float(gamma), q



import lenstronomy.Util.util as Util
import lenstronomy.Util.kernel_util as kernel_util
import lenstronomy.Util.image_util as image_util
import lenstronomy.Util.util as util
from lenstronomy.LightModel.Profiles.gaussian import Gaussian

gaussian = Gaussian()
import pytest
import unittest
import numpy as np
import numpy.testing as npt
from scipy.ndimage import shift


def test_fwhm_kernel():
    x_grid, y_gird = Util.make_grid(101, 1)
    sigma = 20

    flux = gaussian.function(x_grid, y_gird, amp=1, sigma=sigma)
    kernel = Util.array2image(flux)
    kernel = kernel_util.kernel_norm(kernel)
    fwhm_kernel = kernel_util.fwhm_kernel(kernel)
    fwhm = Util.sigma2fwhm(sigma)
    npt.assert_almost_equal(fwhm / fwhm_kernel, 1, 2)


def test_center_kernel():
    x_grid, y_gird = Util.make_grid(31, 1)
    sigma = 2
    flux = gaussian.function(x_grid, y_gird, amp=1, sigma=sigma)
    kernel = Util.array2image(flux)
    kernel = kernel_util.kernel_norm(kernel)

    # kernel being centered
    kernel_new = kernel_util.center_kernel(kernel, iterations=20)
    kernel_new = kernel_util.kernel_norm(kernel_new)
    npt.assert_almost_equal(kernel_new / kernel, 1, decimal=8)
    # kernel shifted in x
    kernel_shifted = shift(kernel, shift=[-0.1, 0], order=1)
    kernel_new = kernel_util.center_kernel(kernel_shifted, iterations=5)
    kernel_new = kernel_util.kernel_norm(kernel_new)
    npt.assert_almost_equal((kernel_new + 0.00001) / (kernel + 0.00001), 1, decimal=4)
    # kernel shifted in y
    kernel_shifted = shift(kernel, shift=[0, -0.4], order=1)
    kernel_new = kernel_util.center_kernel(kernel_shifted, iterations=5)
    kernel_new = kernel_util.kernel_norm(kernel_new)
    npt.assert_almost_equal((kernel_new + 0.01) / (kernel + 0.01), 1, decimal=3)
    # kernel shifted in x and y
    kernel_shifted = shift(kernel, shift=[0.2, -0.3], order=1)
    kernel_new = kernel_util.center_kernel(kernel_shifted, iterations=5)
    kernel_new = kernel_util.kernel_norm(kernel_new)
    npt.assert_almost_equal((kernel_new + 0.01) / (kernel + 0.01), 1, decimal=3)


def test_pixelsize_change():
    kernel = np.zeros((7, 7))
    kernel[3, 3] = 1
    deltaPix_in = 0.1
    deltaPix_out = 0.2
    kernel_new = kernel_util.kernel_pixelsize_change(kernel, deltaPix_in, deltaPix_out)
    assert len(kernel_new) == 3
    assert kernel_new[1, 1] == 1


def test_cutout_source():
    """Test whether a shifted psf can be reproduced sufficiently well :return:"""
    kernel_size = 5
    image = np.zeros((10, 10))
    kernel = np.zeros((kernel_size, kernel_size))
    kernel[2, 2] = 1
    shift_x = 0.5
    shift_y = 0
    x_c, y_c = 5, 5
    x_pos = x_c + shift_x
    y_pos = y_c + shift_y
    # kernel_shifted = interp.shift(kernel, [shift_y, shift_x], order=1)
    image = image_util.add_layer2image(image, x_pos, y_pos, kernel, order=1)
    print(image)
    kernel_new = kernel_util.cutout_source(
        x_pos=x_pos, y_pos=y_pos, image=image, kernelsize=kernel_size
    )
    npt.assert_almost_equal(kernel_new[2, 2], kernel[2, 2], decimal=2)


def test_cutout_source_border():
    kernel_size = 7
    image = np.zeros((10, 10))
    kernel = np.zeros((kernel_size, kernel_size))
    kernel[2, 2] = 1
    shift_x = +0.1
    shift_y = 0
    x_c, y_c = 2, 5
    x_pos = x_c + shift_x
    y_pos = y_c + shift_y
    # kernel_shifted = interp.shift(kernel, [shift_y, shift_x], order=1)
    image = image_util.add_layer2image(image, x_pos, y_pos, kernel, order=1)
    kernel_new = kernel_util.cutout_source(
        x_pos=x_pos, y_pos=y_pos, image=image, kernelsize=kernel_size
    )
    nx_new, ny_new = np.shape(kernel_new)
    print(kernel_new)
    assert nx_new == kernel_size
    assert ny_new == kernel_size
    npt.assert_almost_equal(kernel_new[2, 2], kernel[2, 2], decimal=2)


def test_cut_psf():
    image = np.ones((7, 7))
    psf_cut = kernel_util.cut_psf(image, 5)
    assert len(psf_cut) == 5


def test_de_shift():
    kernel_size = 5
    kernel = np.zeros((kernel_size, kernel_size))
    kernel[2, 2] = 2
    shift_x = 0.48
    shift_y = 0.2
    kernel_shifted = shift(kernel, shift=[-shift_y, -shift_x], order=1)
    kernel_de_shifted = kernel_util.de_shift_kernel(
        kernel_shifted, shift_x, shift_y, iterations=50
    )
    delta_max = np.max(kernel - kernel_de_shifted)
    assert delta_max < 0.01
    npt.assert_almost_equal(kernel_de_shifted[2, 2], kernel[2, 2], decimal=2)

    kernel_size = 5
    kernel = np.zeros((kernel_size, kernel_size))
    kernel[2, 2] = 2
    shift_x = 1.48
    shift_y = 0.2
    kernel_shifted = shift(kernel, shift=[-shift_y, -shift_x], order=1)
    kernel_de_shifted = kernel_util.de_shift_kernel(
        kernel_shifted, shift_x, shift_y, iterations=50
    )
    delta_max = np.max(kernel - kernel_de_shifted)
    assert delta_max < 0.01
    npt.assert_almost_equal(kernel_de_shifted[2, 2], kernel[2, 2], decimal=2)

    kernel_size_x = 5
    kernel_size_y = 4
    kernel = np.zeros((kernel_size_x, kernel_size_y))
    kernel[2, 2] = 2
    shift_x = 1.48
    shift_y = 0.2
    kernel_shifted = shift(kernel, shift=[-shift_y, -shift_x], order=1)
    kernel_de_shifted = kernel_util.de_shift_kernel(
        kernel_shifted, shift_x, shift_y, iterations=50
    )
    delta_max = np.max(kernel - kernel_de_shifted)
    assert delta_max < 0.01
    npt.assert_almost_equal(kernel_de_shifted[2, 2], kernel[2, 2], decimal=2)


def test_deshift_subgrid():
    # test the de-shifting with a sharpened subgrid kernel
    kernel_size = 5
    subgrid = 3
    fwhm = 1
    kernel_subgrid_size = kernel_size * subgrid
    kernel_subgrid = np.zeros((kernel_subgrid_size, kernel_subgrid_size))
    kernel_subgrid[7, 7] = 2
    kernel_subgrid = kernel_util.kernel_gaussian(
        kernel_subgrid_size, 1.0 / subgrid, fwhm=fwhm
    )

    kernel = util.averaging(kernel_subgrid, kernel_subgrid_size, kernel_size)

    shift_x = 0.18
    shift_y = 0.2
    shift_x_subgird = shift_x * subgrid
    shift_y_subgrid = shift_y * subgrid
    kernel_shifted_subgrid = shift(
        kernel_subgrid, shift=[-shift_y_subgrid, -shift_x_subgird], order=1
    )
    kernel_shifted = util.averaging(
        kernel_shifted_subgrid, kernel_subgrid_size, kernel_size
    )
    kernel_shifted_highres = kernel_util.subgrid_kernel(
        kernel_shifted, subgrid_res=subgrid, num_iter=1
    )
    # npt.assert_almost_equal(kernel_shifted_highres[7, 7], kernel_shifted_subgrid[7, 7], decimal=10)


def test_shift_long_dist():
    """Input is a shifted kernel by more than 1 pixel :return:"""

    kernel_size = 9
    kernel = np.zeros((kernel_size, kernel_size))
    kernel[4, 4] = 2.0
    shift_x = 2.0
    shift_y = 1.0
    input_kernel = shift(kernel, shift=[-shift_y, -shift_x], order=1)
    old_style_kernel = shift(input_kernel, shift=[shift_y, shift_x], order=1)
    shifted_new = kernel_util.de_shift_kernel(input_kernel, shift_x, shift_y)
    assert kernel[3, 2] == shifted_new[3, 2]
    assert np.max(old_style_kernel - shifted_new) < 0.01


def test_pixel_kernel():
    # point source kernel
    kernel_size = 9
    kernel = np.zeros((kernel_size, kernel_size))
    kernel[4, 4] = 1.0
    pixel_kernel = kernel_util.pixel_kernel(point_source_kernel=kernel, subgrid_res=1)
    assert pixel_kernel[4, 4] == kernel[4, 4]

    pixel_kernel = kernel_util.pixel_kernel(point_source_kernel=kernel, subgrid_res=11)
    npt.assert_almost_equal(pixel_kernel[4, 4], 0.65841, decimal=3)


def test_split_kernel():
    kernel = np.zeros((9, 9))
    kernel[4, 4] = 1
    subgrid_res = 3
    subgrid_kernel = kernel_util.subgrid_kernel(
        kernel, subgrid_res=subgrid_res, odd=True
    )
    subsampling_size = 3
    kernel_hole, kernel_cutout = kernel_util.split_kernel(
        subgrid_kernel,
        supersampling_kernel_size=subsampling_size,
        supersampling_factor=subgrid_res,
    )

    assert kernel_hole[4, 4] == 0
    assert len(kernel_cutout) == subgrid_res * subsampling_size
    npt.assert_almost_equal(np.sum(kernel_hole) + np.sum(kernel_cutout), 1, decimal=4)

    subgrid_res = 2
    subgrid_kernel = kernel_util.subgrid_kernel(
        kernel, subgrid_res=subgrid_res, odd=True
    )
    subsampling_size = 3
    kernel_hole, kernel_cutout = kernel_util.split_kernel(
        subgrid_kernel,
        supersampling_kernel_size=subsampling_size,
        supersampling_factor=subgrid_res,
    )

    assert kernel_hole[4, 4] == 0
    assert len(kernel_cutout) == subgrid_res * subsampling_size + 1
    npt.assert_almost_equal(np.sum(kernel_hole) + np.sum(kernel_cutout), 1, decimal=4)


def test_cutout_source2():
    grid2d = np.zeros((20, 20))
    grid2d[7:9, 7:9] = 1
    kernel = kernel_util.cutout_source(
        x_pos=7.5, y_pos=7.5, image=grid2d, kernelsize=5, shift=False
    )
    assert kernel[2, 2] == 1


def test_subgrid_kernel():
    kernel = np.zeros((9, 9))
    kernel[4, 4] = 1
    subgrid_res = 3
    subgrid_kernel = kernel_util.subgrid_kernel(
        kernel, subgrid_res=subgrid_res, odd=True
    )
    kernel_re_sized = (
        image_util.re_size(subgrid_kernel, factor=subgrid_res) * subgrid_res**2
    )
    # import matplotlib.pyplot as plt
    # plt.matshow(kernel); plt.show()
    # plt.matshow(subgrid_kernel); plt.show()
    # plt.matshow(kernel_re_sized);plt.show()
    # plt.matshow(kernel_re_sized- kernel);plt.show()
    npt.assert_almost_equal(kernel_re_sized[4, 4], 1, decimal=2)
    assert np.max(subgrid_kernel) == subgrid_kernel[13, 13]
    # assert kernel_re_sized[4, 4] == 1


def test_subgrid_rebin():
    kernel_size = 11
    subgrid_res = 3

    sigma = 1
    x_grid, y_gird = Util.make_grid(kernel_size, 1.0 / subgrid_res, subgrid_res)
    flux = gaussian.function(x_grid, y_gird, amp=1, sigma=sigma)
    kernel = Util.array2image(flux)
    print(np.shape(kernel))
    kernel = util.averaging(
        kernel, numGrid=kernel_size * subgrid_res, numPix=kernel_size
    )
    kernel = kernel_util.kernel_norm(kernel)

    subgrid_kernel = kernel_util.subgrid_kernel(
        kernel, subgrid_res=subgrid_res, odd=True
    )
    kernel_pixel = util.averaging(
        subgrid_kernel, numGrid=kernel_size * subgrid_res, numPix=kernel_size
    )
    kernel_pixel = kernel_util.kernel_norm(kernel_pixel)
    assert np.sum((kernel_pixel - kernel) ** 2) < 0.1


def test_mge_kernel():
    from lenstronomy.LightModel.Profiles.gaussian import MultiGaussian

    mg = MultiGaussian()
    fraction_list = [0.2, 0.7, 0.1]
    sigmas_scaled = [5, 10, 15]
    x, y = util.make_grid(numPix=101, deltapix=1)
    kernel = mg.function(x, y, amp=fraction_list, sigma=sigmas_scaled)
    kernel = util.array2image(kernel)

    amps, sigmas, norm = kernel_util.mge_kernel(kernel, order=10)
    print(amps, sigmas, norm)
    kernel_new = mg.function(x, y, amp=amps, sigma=sigmas)
    kernel_new = util.array2image(kernel_new)
    # npt.assert_almost_equal(sigmas_scaled, sigmas)
    # npt.assert_almost_equal(amps, fraction_list)
    npt.assert_almost_equal(kernel_new, kernel, decimal=3)


def test_kernel_average_pixel():
    gaussian = Gaussian()
    subgrid_res = 3
    x_grid, y_gird = Util.make_grid(9, 1.0, subgrid_res)
    sigma = 2
    flux = gaussian.function(x_grid, y_gird, amp=1, sigma=sigma)
    kernel_super = Util.array2image(flux)
    kernel_pixel = kernel_util.kernel_average_pixel(
        kernel_super, supersampling_factor=subgrid_res
    )
    npt.assert_almost_equal(np.sum(kernel_pixel), np.sum(kernel_super))

    kernel_pixel = kernel_util.kernel_average_pixel(
        kernel_super, supersampling_factor=2
    )
    npt.assert_almost_equal(np.sum(kernel_pixel), np.sum(kernel_super))


def test_averaging_even_kernel():
    subgrid_res = 4

    x_grid, y_gird = Util.make_grid(19, 1.0, 1)
    sigma = 1.5
    flux = gaussian.function(x_grid, y_gird, amp=1, sigma=sigma)
    kernel_super = Util.array2image(flux)

    kernel_pixel = kernel_util.averaging_even_kernel(kernel_super, subgrid_res)
    npt.assert_almost_equal(np.sum(kernel_pixel) * subgrid_res**2, 1, decimal=5)
    assert len(kernel_pixel) == 5

    x_grid, y_gird = Util.make_grid(17, 1.0, 1)
    sigma = 1.5
    amp = 2
    flux = gaussian.function(x_grid, y_gird, amp=amp, sigma=sigma)
    kernel_super = Util.array2image(flux)

    kernel_pixel = kernel_util.averaging_even_kernel(kernel_super, subgrid_res)
    npt.assert_almost_equal(np.sum(kernel_pixel) * subgrid_res**2, amp, decimal=5)
    assert len(kernel_pixel) == 5


def test_degrade_kernel():
    x_grid, y_gird = Util.make_grid(19 * 5, 1.0, 1)
    sigma = 1.5
    amp = 2
    flux = gaussian.function(x_grid, y_gird, amp=2, sigma=sigma)
    kernel_super = Util.array2image(flux) / np.sum(flux) * amp

    for degrading_factor in range(7):
        kernel_degraded = kernel_util.degrade_kernel(
            kernel_super, degrading_factor=degrading_factor + 1
        )
        npt.assert_almost_equal(np.sum(kernel_degraded), amp, decimal=8)


def test_match_kernel_sixe():
    image = np.ones((21, 21))
    size = 11
    image_match = kernel_util.match_kernel_size(image, size)
    nx, ny = np.shape(image_match)
    assert nx == size
    assert ny == size

    image = np.ones((9, 9))
    size = 11
    image_match = kernel_util.match_kernel_size(image, size)
    nx, ny = np.shape(image_match)
    assert nx == size
    assert ny == size

    image = np.ones((11, 11))
    size = 11
    image_match = kernel_util.match_kernel_size(image, size)
    nx, ny = np.shape(image_match)
    assert nx == size
    assert ny == size


def test_kernel_moffat():
    kernel = kernel_util.kernel_moffat(
        num_pix=10, delta_pix=0.1, fwhm=0.5, moffat_beta=1
    )
    npt.assert_almost_equal(np.sum(kernel), 1, decimal=3)


class TestRaise(unittest.TestCase):
    def test_raise(self):
        with self.assertRaises(ValueError):
            kernel = np.zeros((2, 2))
            kernel_util.center_kernel(kernel, iterations=1)

        with self.assertRaises(ValueError):
            kernel_super = np.ones((9, 9))
            kernel_util.split_kernel(
                kernel_super, supersampling_kernel_size=2, supersampling_factor=3
            )
        with self.assertRaises(ValueError):
            kernel_util.split_kernel(
                kernel_super, supersampling_kernel_size=3, supersampling_factor=0
            )
        with self.assertRaises(ValueError):
            image = np.ones((10, 10))
            kernel_util.cutout_source(x_pos=3, y_pos=2, image=image, kernelsize=2)
        with self.assertRaises(ValueError):
            kernel_util.fwhm_kernel(kernel=np.ones((4, 4)))
        with self.assertRaises(ValueError):
            kernel_util.fwhm_kernel(kernel=np.ones((5, 5)))


if __name__ == "__main__":
    pytest.main()



from lenstronomy.LensModel.Profiles.nie import NIE
from lenstronomy.LensModel.Profiles.point_mass import PointMass
from lenstronomy.LensModel.Profiles.base_profile import LensProfileBase
import lenstronomy.Util.param_util as param_util
import numpy as np

from lenstronomy.Util.package_util import exporter

export, __all__ = exporter()


@export
class Chameleon(LensProfileBase):
    """Class of the Chameleon model (See Suyu+2014) an elliptical truncated double
    isothermal profile."""

    param_names = ["alpha_1", "w_c", "w_t", "e1", "e2", "center_x", "center_y"]
    lower_limit_default = {
        "alpha_1": 0,
        "w_c": 0,
        "w_t": 0,
        "e1": -0.8,
        "e2": -0.8,
        "center_x": -100,
        "center_y": -100,
    }
    upper_limit_default = {
        "alpha_1": 100,
        "w_c": 100,
        "w_t": 100,
        "e1": 0.8,
        "e2": 0.8,
        "center_x": 100,
        "center_y": 100,
    }

    def __init__(self, static=False):
        self._nie_1 = NIE()
        self._nie_2 = NIE()
        super(Chameleon, self).__init__()
        self._static = static

    def function(self, x, y, alpha_1, w_c, w_t, e1, e2, center_x=0, center_y=0):
        """

        :param x: ra-coordinate
        :param y: dec-coordinate
        :param alpha_1: deflection angle at 1 (arcseconds) from the center
        :param w_c: see Suyu+2014
        :param w_t: see Suyu+2014
        :param e1: ellipticity parameter
        :param e2: ellipticity parameter
        :param center_x: ra center
        :param center_y: dec center
        :return: lensing potential
        """

        theta_E_conv, w_c, w_t, s_scale_1, s_scale_2 = self.param_convert(
            alpha_1, w_c, w_t, e1, e2
        )
        f_1 = self._nie_1.function(
            x, y, theta_E_conv, e1, e2, s_scale_1, center_x, center_y
        )
        f_2 = self._nie_2.function(
            x, y, theta_E_conv, e1, e2, s_scale_2, center_x, center_y
        )
        f_ = f_1 - f_2
        return f_

    def derivatives(self, x, y, alpha_1, w_c, w_t, e1, e2, center_x=0, center_y=0):
        """

        :param x: ra-coordinate
        :param y: dec-coordinate
        :param alpha_1: deflection angle at 1 (arcseconds) from the center
        :param w_c: see Suyu+2014
        :param w_t: see Suyu+2014
        :param e1: ellipticity parameter
        :param e2: ellipticity parameter
        :param center_x: ra center
        :param center_y: dec center
        :return: deflection angles (RA, DEC)
        """
        theta_E_conv, w_c, w_t, s_scale_1, s_scale_2 = self.param_convert(
            alpha_1, w_c, w_t, e1, e2
        )
        f_x_1, f_y_1 = self._nie_1.derivatives(
            x, y, theta_E_conv, e1, e2, s_scale_1, center_x, center_y
        )
        f_x_2, f_y_2 = self._nie_2.derivatives(
            x, y, theta_E_conv, e1, e2, s_scale_2, center_x, center_y
        )
        f_x = f_x_1 - f_x_2
        f_y = f_y_1 - f_y_2
        return f_x, f_y

    def hessian(self, x, y, alpha_1, w_c, w_t, e1, e2, center_x=0, center_y=0):
        """

        :param x: ra-coordinate
        :param y: dec-coordinate
        :param alpha_1: deflection angle at 1 (arcseconds) from the center
        :param w_c: see Suyu+2014
        :param w_t: see Suyu+2014
        :param e1: ellipticity parameter
        :param e2: ellipticity parameter
        :param center_x: ra center
        :param center_y: dec center
        :return: second derivatives of the lensing potential (Hessian: f_xx, f_xy, f_yx, f_yy)
        """
        theta_E_conv, w_c, w_t, s_scale_1, s_scale_2 = self.param_convert(
            alpha_1, w_c, w_t, e1, e2
        )
        f_xx_1, f_xy_1, f_yx_1, f_yy_1 = self._nie_1.hessian(
            x, y, theta_E_conv, e1, e2, s_scale_1, center_x, center_y
        )
        f_xx_2, f_xy_2, f_yx_2, f_yy_2 = self._nie_2.hessian(
            x, y, theta_E_conv, e1, e2, s_scale_2, center_x, center_y
        )
        f_xx = f_xx_1 - f_xx_2
        f_yy = f_yy_1 - f_yy_2
        f_xy = f_xy_1 - f_xy_2
        f_yx = f_yx_1 - f_yx_2
        return f_xx, f_xy, f_yx, f_yy

    def density_lens(self, r, alpha_1, w_c, w_t, e1=0, e2=0, center_x=0, center_y=0):
        """Spherical average density as a function of 3d radius.

        :param r: 3d radius
        :param alpha_1: deflection angle at 1 (arcseconds) from the center
        :param w_c: see Suyu+2014
        :param w_t: see Suyu+2014
        :param e1: ellipticity parameter
        :param e2: ellipticity parameter
        :param center_x: ra center
        :param center_y: dec center
        :return: matter density at 3d radius r
        """
        theta_E_conv, w_c, w_t, s_scale_1, s_scale_2 = self.param_convert(
            alpha_1, w_c, w_t, e1, e2
        )
        f_1 = self._nie_1.density_lens(
            r, theta_E_conv, e1, e2, s_scale_1, center_x, center_y
        )
        f_2 = self._nie_2.density_lens(
            r, theta_E_conv, e1, e2, s_scale_2, center_x, center_y
        )
        f_ = f_1 - f_2
        return f_

    def mass_3d_lens(self, r, alpha_1, w_c, w_t, e1=0, e2=0, center_x=0, center_y=0):
        """Mass enclosed 3d radius.

        :param r: 3d radius
        :param alpha_1: deflection angle at 1 (arcseconds) from the center
        :param w_c: see Suyu+2014
        :param w_t: see Suyu+2014
        :param e1: ellipticity parameter
        :param e2: ellipticity parameter
        :param center_x: ra center
        :param center_y: dec center
        :return: mass enclosed 3d radius r
        """
        theta_E_conv, w_c, w_t, s_scale_1, s_scale_2 = self.param_convert(
            alpha_1, w_c, w_t, e1, e2
        )
        m_1 = self._nie_1.mass_3d_lens(
            r, theta_E_conv, e1, e2, s_scale_1, center_x, center_y
        )
        m_2 = self._nie_2.mass_3d_lens(
            r, theta_E_conv, e1, e2, s_scale_2, center_x, center_y
        )
        m_ = m_1 - m_2
        return m_

    def param_convert(self, alpha_1, w_c, w_t, e1, e2):
        """Convert the parameter alpha_1 (deflection angle one arcsecond from the
        center) into the "Einstein radius" scale parameter of the two NIE profiles.

        :param alpha_1: deflection angle at 1 (arcseconds) from the center
        :param w_c: see Suyu+2014
        :param w_t: see Suyu+2014
        :param e1: eccentricity modulus
        :param e2: eccentricity modulus
        :return:
        """
        if self._static is True:
            return (
                self._theta_convert_static,
                self._w_c_static,
                self._w_t_stactic,
                self._s_scale_1_static,
                self._s_scale_2_static,
            )
        return self._param_convert(alpha_1, w_c, w_t, e1, e2)

    def _param_convert(self, alpha_1, w_c, w_t, e1, e2):
        if not w_t >= w_c:
            return 0, w_t, w_c, 1, 1
        s_scale_1 = w_c
        s_scale_2 = w_t
        f_x_1, f_y_1 = self._nie_1.derivatives(
            1, 0, theta_E=1, e1=0, e2=0, s_scale=s_scale_1
        )
        f_x_2, f_y_2 = self._nie_2.derivatives(
            1, 0, theta_E=1, e1=0, e2=0, s_scale=s_scale_2
        )
        f_x = f_x_1 - f_x_2
        theta_E_convert = alpha_1 / f_x
        phi_G, q = param_util.ellipticity2phi_q(e1, e2)
        # TODO: is this next conversion really needed since the NIE definition is already in the average sense?
        s_scale_1 = np.sqrt(4 * w_c**2 / (1.0 + q) ** 2)
        s_scale_2 = np.sqrt(4 * w_t**2 / (1.0 + q) ** 2)
        return theta_E_convert, w_c, w_t, s_scale_1, s_scale_2

    def set_static(self, alpha_1, w_c, w_t, e1, e2, center_x=0, center_y=0):
        """

        :param alpha_1:
        :param w_c:
        :param w_t:
        :param e1:
        :param e2:
        :param center_x:
        :param center_y:
        :return:
        """
        self._static = True
        (
            self._theta_convert_static,
            self._w_c_static,
            self._w_t_stactic,
            self._s_scale_1_static,
            self._s_scale_2_static,
        ) = self._param_convert(alpha_1, w_c, w_t, e1, e2)
        self._nie_1.set_static(
            self._theta_convert_static,
            e1,
            e2,
            self._s_scale_1_static,
            center_x,
            center_y,
        )
        self._nie_2.set_static(
            self._theta_convert_static,
            e1,
            e2,
            self._s_scale_2_static,
            center_x,
            center_y,
        )

    def set_dynamic(self):
        """

        :return:
        """
        self._static = False
        if hasattr(self, "_theta_convert_static"):
            del self._theta_convert_static
        if hasattr(self, "_w_c_static"):
            del self._w_c_static
        if hasattr(self, "_w_t_stactic"):
            del self._w_t_stactic
        if hasattr(self, "_s_scale_1_static"):
            del self._s_scale_1_static
        if hasattr(self, "_s_scale_2_static"):
            del self._s_scale_2_static
        self._nie_1.set_dynamic()
        self._nie_2.set_dynamic()


@export
class DoubleChameleon(LensProfileBase):
    """Class of the Chameleon model (See Suyu+2014) an elliptical truncated double
    isothermal profile."""

    param_names = [
        "alpha_1",
        "ratio",
        "w_c1",
        "w_t1",
        "e11",
        "e21",
        "w_c2",
        "w_t2",
        "e12",
        "e22",
        "center_x",
        "center_y",
    ]
    lower_limit_default = {
        "alpha_1": 0,
        "ratio": 0,
        "w_c1": 0,
        "w_t1": 0,
        "e11": -0.8,
        "e21": -0.8,
        "w_c2": 0,
        "w_t2": 0,
        "e12": -0.8,
        "e22": -0.8,
        "center_x": -100,
        "center_y": -100,
    }
    upper_limit_default = {
        "alpha_1": 100,
        "ratio": 100,
        "w_c1": 100,
        "w_t1": 100,
        "e11": 0.8,
        "e21": 0.8,
        "w_c2": 100,
        "w_t2": 100,
        "e12": 0.8,
        "e22": 0.8,
        "center_x": 100,
        "center_y": 100,
    }

    def __init__(self):
        self._chameleon_1 = Chameleon()
        self._chameleon_2 = Chameleon()
        super(DoubleChameleon, self).__init__()

    def function(
        self,
        x,
        y,
        alpha_1,
        ratio,
        w_c1,
        w_t1,
        e11,
        e21,
        w_c2,
        w_t2,
        e12,
        e22,
        center_x=0,
        center_y=0,
    ):
        """
        :param x: ra-coordinate
        :param y: dec-coordinate
        :param alpha_1: deflection angle at 1 (arcseconds) from the center
        :param ratio: ratio of deflection amplitude at radius = 1 of the first to second Chameleon profile
        :param w_c1: Suyu+2014 for first profile
        :param w_t1: Suyu+2014 for first profile
        :param e11: ellipticity parameter for first profile
        :param e21: ellipticity parameter for first profile
        :param w_c2: Suyu+2014 for second profile
        :param w_t2: Suyu+2014 for second profile
        :param e12: ellipticity parameter for second profile
        :param e22: ellipticity parameter for second profile
        :param center_x: ra center
        :param center_y: dec center
        :return: lensing potential
        """

        f_1 = self._chameleon_1.function(
            x,
            y,
            alpha_1 / (1.0 + 1.0 / ratio),
            w_c1,
            w_t1,
            e11,
            e21,
            center_x,
            center_y,
        )
        f_2 = self._chameleon_2.function(
            x, y, alpha_1 / (1.0 + ratio), w_c2, w_t2, e12, e22, center_x, center_y
        )
        return f_1 + f_2

    def derivatives(
        self,
        x,
        y,
        alpha_1,
        ratio,
        w_c1,
        w_t1,
        e11,
        e21,
        w_c2,
        w_t2,
        e12,
        e22,
        center_x=0,
        center_y=0,
    ):
        """
        :param x: ra-coordinate
        :param y: dec-coordinate
        :param alpha_1: deflection angle at 1 (arcseconds) from the center
        :param ratio: ratio of deflection amplitude at radius = 1 of the first to second Chameleon profile
        :param w_c1: Suyu+2014 for first profile
        :param w_t1: Suyu+2014 for first profile
        :param e11: ellipticity parameter for first profile
        :param e21: ellipticity parameter for first profile
        :param w_c2: Suyu+2014 for second profile
        :param w_t2: Suyu+2014 for second profile
        :param e12: ellipticity parameter for second profile
        :param e22: ellipticity parameter for second profile^V
        :param center_x: ra center
        :param center_y: dec center
        :return: deflection angles (RA, DEC)
        """
        f_x1, f_y1 = self._chameleon_1.derivatives(
            x,
            y,
            alpha_1 / (1.0 + 1.0 / ratio),
            w_c1,
            w_t1,
            e11,
            e21,
            center_x,
            center_y,
        )
        f_x2, f_y2 = self._chameleon_2.derivatives(
            x, y, alpha_1 / (1.0 + ratio), w_c2, w_t2, e12, e22, center_x, center_y
        )
        return f_x1 + f_x2, f_y1 + f_y2

    def hessian(
        self,
        x,
        y,
        alpha_1,
        ratio,
        w_c1,
        w_t1,
        e11,
        e21,
        w_c2,
        w_t2,
        e12,
        e22,
        center_x=0,
        center_y=0,
    ):
        """
        :param x: ra-coordinate
        :param y: dec-coordinate
        :param alpha_1: deflection angle at 1 (arcseconds) from the center
        :param ratio: ratio of deflection amplitude at radius = 1 of the first to second Chameleon profile
        :param w_c1: Suyu+2014 for first profile
        :param w_t1: Suyu+2014 for first profile
        :param e11: ellipticity parameter for first profile
        :param e21: ellipticity parameter for first profile
        :param w_c2: Suyu+2014 for second profile
        :param w_t2: Suyu+2014 for second profile
        :param e12: ellipticity parameter for second profile
        :param e22: ellipticity parameter for second profile
        :param center_x: ra center
        :param center_y: dec center
        :return: second derivatives of the lensing potential (Hessian: f_xx, f_yy, f_xy)
        """
        (
            f_xx1,
            f_xy1,
            f_yx1,
            f_yy1,
        ) = self._chameleon_1.hessian(
            x,
            y,
            alpha_1 / (1.0 + 1.0 / ratio),
            w_c1,
            w_t1,
            e11,
            e21,
            center_x,
            center_y,
        )
        f_xx2, f_xy2, f_yx2, f_yy2 = self._chameleon_2.hessian(
            x, y, alpha_1 / (1.0 + ratio), w_c2, w_t2, e12, e22, center_x, center_y
        )
        return f_xx1 + f_xx2, f_xy1 + f_xy2, f_xy1 + f_xy2, f_yy1 + f_yy2

    def density_lens(
        self,
        r,
        alpha_1,
        ratio,
        w_c1,
        w_t1,
        e11,
        e21,
        w_c2,
        w_t2,
        e12,
        e22,
        center_x=0,
        center_y=0,
    ):
        """
        :param r: 3d radius
        :param alpha_1: deflection angle at 1 (arcseconds) from the center
        :param ratio: ratio of deflection amplitude at radius = 1 of the first to second Chameleon profile
        :param w_c1: Suyu+2014 for first profile
        :param w_t1: Suyu+2014 for first profile
        :param e11: ellipticity parameter for first profile
        :param e21: ellipticity parameter for first profile
        :param w_c2: Suyu+2014 for second profile
        :param w_t2: Suyu+2014 for second profile
        :param e12: ellipticity parameter for second profile
        :param e22: ellipticity parameter for second profile
        :param center_x: ra center
        :param center_y: dec center
        :return: 3d density at radius r
        """

        f_1 = self._chameleon_1.density_lens(
            r, alpha_1 / (1.0 + 1.0 / ratio), w_c1, w_t1, e11, e21, center_x, center_y
        )
        f_2 = self._chameleon_2.density_lens(
            r, alpha_1 / (1.0 + ratio), w_c2, w_t2, e12, e22, center_x, center_y
        )
        return f_1 + f_2

    def mass_3d_lens(
        self,
        r,
        alpha_1,
        ratio,
        w_c1,
        w_t1,
        e11,
        e21,
        w_c2,
        w_t2,
        e12,
        e22,
        center_x=0,
        center_y=0,
    ):
        """
        :param r: 3d radius
        :param alpha_1: deflection angle at 1 (arcseconds) from the center
        :param ratio: ratio of deflection amplitude at radius = 1 of the first to second Chameleon profile
        :param w_c1: Suyu+2014 for first profile
        :param w_t1: Suyu+2014 for first profile
        :param e11: ellipticity parameter for first profile
        :param e21: ellipticity parameter for first profile
        :param w_c2: Suyu+2014 for second profile
        :param w_t2: Suyu+2014 for second profile
        :param e12: ellipticity parameter for second profile
        :param e22: ellipticity parameter for second profile
        :param center_x: ra center
        :param center_y: dec center
        :return: mass enclosed 3d radius
        """

        m_1 = self._chameleon_1.mass_3d_lens(
            r, alpha_1 / (1.0 + 1.0 / ratio), w_c1, w_t1, e11, e21, center_x, center_y
        )
        m_2 = self._chameleon_2.mass_3d_lens(
            r, alpha_1 / (1.0 + ratio), w_c2, w_t2, e12, e22, center_x, center_y
        )
        return m_1 + m_2

    def set_static(
        self,
        alpha_1,
        ratio,
        w_c1,
        w_t1,
        e11,
        e21,
        w_c2,
        w_t2,
        e12,
        e22,
        center_x=0,
        center_y=0,
    ):
        self._chameleon_1.set_static(
            alpha_1 / (1.0 + 1.0 / ratio), w_c1, w_t1, e11, e21, center_x, center_y
        )
        self._chameleon_2.set_static(
            alpha_1 / (1.0 + ratio), w_c2, w_t2, e12, e22, center_x, center_y
        )

    def set_dynamic(self):
        self._chameleon_1.set_dynamic()
        self._chameleon_2.set_dynamic()


@export
class TripleChameleon(LensProfileBase):
    """Class of the Chameleon model (See Suyu+2014) an elliptical truncated double
    isothermal profile."""

    param_names = [
        "alpha_1",
        "ratio12",
        "ratio13",
        "w_c1",
        "w_t1",
        "e11",
        "e21",
        "w_c2",
        "w_t2",
        "e12",
        "e22",
        "w_c3",
        "w_t3",
        "e13",
        "e23",
        "center_x",
        "center_y",
    ]
    lower_limit_default = {
        "alpha_1": 0,
        "ratio12": 0,
        "ratio13": 0,
        "w_c1": 0,
        "w_t1": 0,
        "e11": -0.8,
        "e21": -0.8,
        "w_c2": 0,
        "w_t2": 0,
        "e12": -0.8,
        "e22": -0.8,
        "w_c3": 0,
        "w_t3": 0,
        "e13": -0.8,
        "e23": -0.8,
        "center_x": -100,
        "center_y": -100,
    }
    upper_limit_default = {
        "alpha_1": 100,
        "ratio12": 100,
        "ratio13": 100,
        "w_c1": 100,
        "w_t1": 100,
        "e11": 0.8,
        "e21": 0.8,
        "w_c2": 100,
        "w_t2": 100,
        "e12": 0.8,
        "e22": 0.8,
        "w_c3": 100,
        "w_t3": 100,
        "e13": 0.8,
        "e23": 0.8,
        "center_x": 100,
        "center_y": 100,
    }

    def __init__(self):
        self._chameleon_1 = Chameleon()
        self._chameleon_2 = Chameleon()
        self._chameleon_3 = Chameleon()
        super(TripleChameleon, self).__init__()

    @staticmethod
    def _ratio_definition(alpha_1, ratio12, ratio13):
        """

        :param alpha_1: deflection angle at 1 arcsecond
        :param ratio12: ratio of first to second amplitude
        :param ratio13: ratio of first to third amplitude
        :return: amplitudes of individual chameleon profiles
        """
        amp1 = alpha_1 / (1.0 + 1.0 / ratio12 + 1.0 / ratio13)
        amp2 = amp1 / ratio12
        amp3 = amp1 / ratio13
        return amp1, amp2, amp3

    def function(
        self,
        x,
        y,
        alpha_1,
        ratio12,
        ratio13,
        w_c1,
        w_t1,
        e11,
        e21,
        w_c2,
        w_t2,
        e12,
        e22,
        w_c3,
        w_t3,
        e13,
        e23,
        center_x=0,
        center_y=0,
    ):
        """

        :param alpha_1:
        :param ratio12: ratio of first to second amplitude
        :param ratio13: ratio of first to third amplitude
        :param w_c1:
        :param w_t1:
        :param e11:
        :param e21:
        :param w_c2:
        :param w_t2:
        :param e12:
        :param e22:
        :param center_x:
        :param center_y:
        :return:
        """
        amp1, amp2, amp3 = self._ratio_definition(alpha_1, ratio12, ratio13)
        f_1 = self._chameleon_1.function(
            x, y, amp1, w_c1, w_t1, e11, e21, center_x, center_y
        )
        f_2 = self._chameleon_2.function(
            x, y, amp2, w_c2, w_t2, e12, e22, center_x, center_y
        )
        f_3 = self._chameleon_3.function(
            x, y, amp3, w_c3, w_t3, e13, e23, center_x, center_y
        )
        return f_1 + f_2 + f_3

    def derivatives(
        self,
        x,
        y,
        alpha_1,
        ratio12,
        ratio13,
        w_c1,
        w_t1,
        e11,
        e21,
        w_c2,
        w_t2,
        e12,
        e22,
        w_c3,
        w_t3,
        e13,
        e23,
        center_x=0,
        center_y=0,
    ):
        """

        :param alpha_1:
        :param ratio12: ratio of first to second amplitude
        :param ratio13: ratio of first to third amplidute
        :param w_c1:
        :param w_t1:
        :param e11:
        :param e21:
        :param w_c2:
        :param w_t2:
        :param e12:
        :param e22:
        :param center_x:
        :param center_y:
        :return:
        """
        amp1, amp2, amp3 = self._ratio_definition(alpha_1, ratio12, ratio13)
        f_x1, f_y1 = self._chameleon_1.derivatives(
            x, y, amp1, w_c1, w_t1, e11, e21, center_x, center_y
        )
        f_x2, f_y2 = self._chameleon_2.derivatives(
            x, y, amp2, w_c2, w_t2, e12, e22, center_x, center_y
        )
        f_x3, f_y3 = self._chameleon_3.derivatives(
            x, y, amp3, w_c3, w_t3, e13, e23, center_x, center_y
        )
        return f_x1 + f_x2 + f_x3, f_y1 + f_y2 + f_y3

    def hessian(
        self,
        x,
        y,
        alpha_1,
        ratio12,
        ratio13,
        w_c1,
        w_t1,
        e11,
        e21,
        w_c2,
        w_t2,
        e12,
        e22,
        w_c3,
        w_t3,
        e13,
        e23,
        center_x=0,
        center_y=0,
    ):
        """

        :param alpha_1:
        :param ratio12: ratio of first to second amplitude
        :param ratio13: ratio of first to third amplidute
        :param w_c1:
        :param w_t1:
        :param e11:
        :param e21:
        :param w_c2:
        :param w_t2:
        :param e12:
        :param e22:
        :param center_x:
        :param center_y:
        :return:
        """
        amp1, amp2, amp3 = self._ratio_definition(alpha_1, ratio12, ratio13)
        f_xx1, f_xy1, f_yx1, f_yy1 = self._chameleon_1.hessian(
            x, y, amp1, w_c1, w_t1, e11, e21, center_x, center_y
        )
        f_xx2, f_xy2, f_yx2, f_yy2 = self._chameleon_2.hessian(
            x, y, amp2, w_c2, w_t2, e12, e22, center_x, center_y
        )
        f_xx3, f_xy3, f_yx3, f_yy3 = self._chameleon_3.hessian(
            x, y, amp3, w_c3, w_t3, e13, e23, center_x, center_y
        )
        return (
            f_xx1 + f_xx2 + f_xx3,
            f_xy1 + f_xy2 + f_xy3,
            f_yx1 + f_yx2 + f_yx3,
            f_yy1 + f_yy2 + f_yy3,
        )

    def density_lens(
        self,
        r,
        alpha_1,
        ratio12,
        ratio13,
        w_c1,
        w_t1,
        e11,
        e21,
        w_c2,
        w_t2,
        e12,
        e22,
        w_c3,
        w_t3,
        e13,
        e23,
        center_x=0,
        center_y=0,
    ):
        """

        :param r: 3d radius
        :param alpha_1:
        :param ratio12: ratio of first to second amplitude
        :param ratio13: ratio of first to third amplitude
        :param w_c1:
        :param w_t1:
        :param e11:
        :param e21:
        :param w_c2:
        :param w_t2:
        :param e12:
        :param e22:
        :param center_x:
        :param center_y:
        :return: density at radius r (spherical average)
        """
        amp1, amp2, amp3 = self._ratio_definition(alpha_1, ratio12, ratio13)
        f_1 = self._chameleon_1.density_lens(
            r, amp1, w_c1, w_t1, e11, e21, center_x, center_y
        )
        f_2 = self._chameleon_2.density_lens(
            r, amp2, w_c2, w_t2, e12, e22, center_x, center_y
        )
        f_3 = self._chameleon_3.density_lens(
            r, amp3, w_c3, w_t3, e13, e23, center_x, center_y
        )
        return f_1 + f_2 + f_3

    def mass_3d_lens(
        self,
        r,
        alpha_1,
        ratio12,
        ratio13,
        w_c1,
        w_t1,
        e11,
        e21,
        w_c2,
        w_t2,
        e12,
        e22,
        w_c3,
        w_t3,
        e13,
        e23,
        center_x=0,
        center_y=0,
    ):
        """

        :param r: 3d radius
        :param alpha_1:
        :param ratio12: ratio of first to second amplitude
        :param ratio13: ratio of first to third amplitude
        :param w_c1:
        :param w_t1:
        :param e11:
        :param e21:
        :param w_c2:
        :param w_t2:
        :param e12:
        :param e22:
        :param center_x:
        :param center_y:
        :return: mass enclosed 3d radius
        """
        amp1, amp2, amp3 = self._ratio_definition(alpha_1, ratio12, ratio13)
        m_1 = self._chameleon_1.mass_3d_lens(
            r, amp1, w_c1, w_t1, e11, e21, center_x, center_y
        )
        m_2 = self._chameleon_2.mass_3d_lens(
            r, amp2, w_c2, w_t2, e12, e22, center_x, center_y
        )
        m_3 = self._chameleon_3.mass_3d_lens(
            r, amp3, w_c3, w_t3, e13, e23, center_x, center_y
        )
        return m_1 + m_2 + m_3

    def set_static(
        self,
        alpha_1,
        ratio12,
        ratio13,
        w_c1,
        w_t1,
        e11,
        e21,
        w_c2,
        w_t2,
        e12,
        e22,
        w_c3,
        w_t3,
        e13,
        e23,
        center_x=0,
        center_y=0,
    ):
        amp1, amp2, amp3 = self._ratio_definition(alpha_1, ratio12, ratio13)
        self._chameleon_1.set_static(amp1, w_c1, w_t1, e11, e21, center_x, center_y)
        self._chameleon_2.set_static(amp2, w_c2, w_t2, e12, e22, center_x, center_y)
        self._chameleon_3.set_static(amp3, w_c3, w_t3, e13, e23, center_x, center_y)

    def set_dynamic(self):
        self._chameleon_1.set_dynamic()
        self._chameleon_2.set_dynamic()
        self._chameleon_3.set_dynamic()


@export
class DoubleChameleonPointMass(LensProfileBase):
    """Class of the Chameleon model (See Suyu+2014) an elliptical truncated double
    isothermal profile."""

    param_names = [
        "alpha_1",
        "ratio_chameleon",
        "ratio_pointmass",
        "w_c1",
        "w_t1",
        "e11",
        "e21",
        "w_c2",
        "w_t2",
        "e12",
        "e22",
        "center_x",
        "center_y",
    ]
    lower_limit_default = {
        "alpha_1": 0,
        "ratio_chameleon": 0,
        "ratio_pointmass": 0,
        "w_c1": 0,
        "w_t1": 0,
        "e11": -0.8,
        "e21": -0.8,
        "w_c2": 0,
        "w_t2": 0,
        "e12": -0.8,
        "e22": -0.8,
        "center_x": -100,
        "center_y": -100,
    }
    upper_limit_default = {
        "alpha_1": 100,
        "ratio_chameleon": 100,
        "ratio_pointmass": 100,
        "w_c1": 100,
        "w_t1": 100,
        "e11": 0.8,
        "e21": 0.8,
        "w_c2": 100,
        "w_t2": 100,
        "e12": 0.8,
        "e22": 0.8,
        "center_x": 100,
        "center_y": 100,
    }

    def __init__(self):
        self.chameleon = DoubleChameleon()
        self.pointMass = PointMass()
        super(DoubleChameleonPointMass, self).__init__()

    def function(
        self,
        x,
        y,
        alpha_1,
        ratio_pointmass,
        ratio_chameleon,
        w_c1,
        w_t1,
        e11,
        e21,
        w_c2,
        w_t2,
        e12,
        e22,
        center_x=0,
        center_y=0,
    ):
        """#TODO chose better parameterization for combining point mass and Chameleon
        profiles

        :param x: ra-coordinate
        :param y: dec-coordinate
        :param alpha_1: deflection angle at 1 (arcseconds) from the center
        :param ratio_pointmass: ratio of point source Einstein radius to combined
            Chameleon deflection angle at r=1
        :param ratio_chameleon: ratio in deflection angles at r=1 for the two Chameleon
            profiles
        :param w_c1: Suyu+2014 for first profile
        :param w_t1: Suyu+2014 for first profile
        :param e11: ellipticity parameter for first profile
        :param e21: ellipticity parameter for first profile
        :param w_c2: Suyu+2014 for second profile
        :param w_t2: Suyu+2014 for second profile
        :param e12: ellipticity parameter for second profile
        :param e22: ellipticity parameter for second profile
        :param center_x: ra center
        :param center_y: dec center
        :return: lensing potential
        """
        f_1 = self.pointMass.function(
            x, y, alpha_1 / (1.0 + 1.0 / ratio_pointmass), center_x, center_y
        )
        f_2 = self.chameleon.function(
            x,
            y,
            alpha_1 / (1.0 + ratio_pointmass),
            ratio_chameleon,
            w_c1,
            w_t1,
            e11,
            e21,
            w_c2,
            w_t2,
            e12,
            e22,
            center_x,
            center_y,
        )
        return f_1 + f_2

    def derivatives(
        self,
        x,
        y,
        alpha_1,
        ratio_pointmass,
        ratio_chameleon,
        w_c1,
        w_t1,
        e11,
        e21,
        w_c2,
        w_t2,
        e12,
        e22,
        center_x=0,
        center_y=0,
    ):
        """

        :param x:
        :param y:
        :param alpha_1:
        :param ratio_pointmass: ratio of point source Einstein radius to combined Chameleon deflection angle at r=1
        :param ratio_chameleon: ratio in deflection angles at r=1 for the two Chameleon profiles
        :param w_c1: Suyu+2014 for first profile
        :param w_t1: Suyu+2014 for first profile
        :param e11: ellipticity parameter for first profile
        :param e21: ellipticity parameter for first profile
        :param w_c2: Suyu+2014 for second profile
        :param w_t2: Suyu+2014 for second profile
        :param e12: ellipticity parameter for second profile
        :param e22: ellipticity parameter for second profile
        :param center_x: ra center
        :param center_y: dec center
        :return:
        """
        f_x1, f_y1 = self.pointMass.derivatives(
            x, y, alpha_1 / (1.0 + 1.0 / ratio_pointmass), center_x, center_y
        )
        f_x2, f_y2 = self.chameleon.derivatives(
            x,
            y,
            alpha_1 / (1.0 + ratio_pointmass),
            ratio_chameleon,
            w_c1,
            w_t1,
            e11,
            e21,
            w_c2,
            w_t2,
            e12,
            e22,
            center_x,
            center_y,
        )
        return f_x1 + f_x2, f_y1 + f_y2

    def hessian(
        self,
        x,
        y,
        alpha_1,
        ratio_pointmass,
        ratio_chameleon,
        w_c1,
        w_t1,
        e11,
        e21,
        w_c2,
        w_t2,
        e12,
        e22,
        center_x=0,
        center_y=0,
    ):
        """

        :param x:
        :param y:
        :param alpha_1:
        :param ratio_pointmass: ratio of point source Einstein radius to combined Chameleon deflection angle at r=1
        :param ratio_chameleon: ratio in deflection angles at r=1 for the two Chameleon profiles
        :param w_c1: Suyu+2014 for first profile
        :param w_t1: Suyu+2014 for first profile
        :param e11: ellipticity parameter for first profile
        :param e21: ellipticity parameter for first profile
        :param w_c2: Suyu+2014 for second profile
        :param w_t2: Suyu+2014 for second profile
        :param e12: ellipticity parameter for second profile
        :param e22: ellipticity parameter for second profile
        :param center_x: ra center
        :param center_y: dec center
        :return:
        """
        f_xx1, f_xy1, f_yx1, f_yy1 = self.pointMass.hessian(
            x, y, alpha_1 / (1.0 + 1.0 / ratio_pointmass), center_x, center_y
        )
        f_xx2, f_xy2, f_yx2, f_yy2 = self.chameleon.hessian(
            x,
            y,
            alpha_1 / (1.0 + ratio_pointmass),
            ratio_chameleon,
            w_c1,
            w_t1,
            e11,
            e21,
            w_c2,
            w_t2,
            e12,
            e22,
            center_x,
            center_y,
        )
        return f_xx1 + f_xx2, f_xy1 + f_xy2, f_yx1 + f_yx2, f_yy1 + f_yy2



__author__ = "sibirrer"

import pytest
import numpy as np
import lenstronomy.Util.simulation_util as sim_util
from lenstronomy.ImSim.image_model import ImageModel
import lenstronomy.Util.param_util as param_util
from lenstronomy.PointSource.point_source import PointSource
from lenstronomy.LensModel.lens_model import LensModel
from lenstronomy.LightModel.light_model import LightModel
from lenstronomy.Plots.model_plot import ModelPlot
from lenstronomy.Plots.model_plot import check_solver_error
from lenstronomy.Data.imaging_data import ImageData
from lenstronomy.Data.psf import PSF

import matplotlib

matplotlib.use("agg")
import matplotlib.pyplot as plt
import unittest


class TestOutputPlots(object):
    """Test the fitting sequences."""

    def setup_method(self):
        # data specifics
        sigma_bkg = 0.05  # background noise per pixel
        exp_time = 100  # exposure time (arbitrary units, flux per pixel is in units #photons/exp_time unit)
        numPix = 10  # cutout pixel size
        deltaPix = 0.5  # pixel size in arcsec (area per pixel = deltaPix**2)
        fwhm = 0.5  # full width half max of PSF

        # PSF specification

        self.kwargs_data = sim_util.data_configure_simple(
            numPix, deltaPix, exp_time, sigma_bkg
        )
        data_class = ImageData(**self.kwargs_data)
        kwargs_psf_gaussian = {
            "psf_type": "GAUSSIAN",
            "fwhm": fwhm,
            "truncation": 5,
            "pixel_size": deltaPix,
        }
        psf_gaussian = PSF(**kwargs_psf_gaussian)
        self.kwargs_psf = {
            "psf_type": "PIXEL",
            "kernel_point_source": psf_gaussian.kernel_point_source,
        }
        psf_class = PSF(**self.kwargs_psf)

        # 'EXERNAL_SHEAR': external shear
        kwargs_shear = {
            "gamma1": 0.01,
            "gamma2": 0.01,
        }  # gamma_ext: shear strength, psi_ext: shear angel (in radian)
        e1, e2 = param_util.phi_q2_ellipticity(0.2, 0.8)
        kwargs_spemd = {
            "theta_E": 1.0,
            "gamma": 1.8,
            "center_x": 0,
            "center_y": 0,
            "e1": e1,
            "e2": e2,
        }

        lens_model_list = ["SPEP", "SHEAR"]
        self.kwargs_lens = [kwargs_spemd, kwargs_shear]
        lens_model_class = LensModel(
            lens_model_list=lens_model_list,
            multi_plane=True,
            lens_redshift_list=[0.5, 0.5],
            z_source=2.0,
        )
        self.LensModel = lens_model_class

        # list of light profiles (for lens and source)
        # 'SERSIC': spherical Sersic profile
        kwargs_sersic = {
            "amp": 1.0,
            "R_sersic": 0.1,
            "n_sersic": 2,
            "center_x": 0,
            "center_y": 0,
        }
        # 'SERSIC_ELLIPSE': elliptical Sersic profile
        phi, q = 0.2, 0.9
        e1, e2 = param_util.phi_q2_ellipticity(phi, q)
        kwargs_sersic_ellipse = {
            "amp": 1.0,
            "R_sersic": 0.6,
            "n_sersic": 7,
            "center_x": 0,
            "center_y": 0,
            "e1": e1,
            "e2": e2,
        }

        lens_light_model_list = ["SERSIC"]
        self.kwargs_lens_light = [kwargs_sersic]
        lens_light_model_class = LightModel(light_model_list=lens_light_model_list)
        source_model_list = ["SERSIC_ELLIPSE"]
        self.kwargs_source = [kwargs_sersic_ellipse]
        source_model_class = LightModel(light_model_list=source_model_list)
        self.kwargs_ps = [
            {"ra_source": 0.0, "dec_source": 0.0, "source_amp": 1.0}
        ]  # quasar point source position in the source plane and intrinsic brightness
        point_source_list = ["SOURCE_POSITION"]
        point_source_class = PointSource(
            point_source_type_list=point_source_list, fixed_magnification_list=[True]
        )
        kwargs_numerics = {"supersampling_factor": 1}
        imageModel = ImageModel(
            data_class,
            psf_class,
            lens_model_class,
            source_model_class,
            lens_light_model_class,
            point_source_class,
            kwargs_numerics=kwargs_numerics,
        )
        image_sim = sim_util.simulate_simple(
            imageModel,
            self.kwargs_lens,
            self.kwargs_source,
            self.kwargs_lens_light,
            self.kwargs_ps,
        )

        data_class.update_data(image_sim)
        self.kwargs_data["image_data"] = image_sim
        self.kwargs_model = {
            "lens_model_list": lens_model_list,
            "source_light_model_list": source_model_list,
            "lens_light_model_list": lens_light_model_list,
            "point_source_model_list": point_source_list,
            "fixed_magnification_list": [False],
        }
        self.kwargs_model_multiplane = {
            "lens_model_list": lens_model_list,
            "lens_redshift_list": [0.5, 0.5],
            "multi_plane": True,
            "z_source": 2.0,
            "source_light_model_list": source_model_list,
            "lens_light_model_list": lens_light_model_list,
            "point_source_model_list": point_source_list,
            "fixed_magnification_list": [False],
        }
        self.kwargs_numerics = kwargs_numerics
        self.data_class = ImageData(**self.kwargs_data)
        self.kwargs_params = {
            "kwargs_lens": self.kwargs_lens,
            "kwargs_source": self.kwargs_source,
            "kwargs_lens_light": self.kwargs_lens_light,
            "kwargs_ps": self.kwargs_ps,
        }

    def test_lensModelPlot(self):
        multi_band_list = [[self.kwargs_data, self.kwargs_psf, self.kwargs_numerics]]
        lensPlot = ModelPlot(
            multi_band_list,
            self.kwargs_model,
            self.kwargs_params,
            arrow_size=0.02,
            cmap_string="gist_heat",
            multi_band_type="single-band",
        )

        multi_band_list_multiplane = [
            [self.kwargs_data, self.kwargs_psf, self.kwargs_numerics]
        ]
        lensPlot_multiplane = ModelPlot(
            multi_band_list_multiplane,
            self.kwargs_model_multiplane,
            self.kwargs_params,
            arrow_size=0.02,
            cmap_string="gist_heat",
            multi_band_type="single-band",
        )

        lensPlot.plot_main(with_caustics=True)
        plt.close()
        cmap = plt.get_cmap("gist_heat")

        lensPlot = ModelPlot(
            multi_band_list,
            self.kwargs_model,
            self.kwargs_params,
            arrow_size=0.02,
            cmap_string=cmap,
        )

        lensPlot.plot_separate()
        plt.close()
        lensPlot.plot_subtract_from_data_all()
        plt.close()
        f, ax = plt.subplots(1, 1, figsize=(4, 4))
        lensPlot.deflection_plot(ax=ax, with_caustics=True, axis=1)
        plt.close()

        f, ax = plt.subplots(1, 1, figsize=(4, 4))
        lensPlot.subtract_from_data_plot(ax=ax)
        plt.close()

        f, ax = plt.subplots(1, 1, figsize=(4, 4))
        lensPlot.deflection_plot(ax=ax, with_caustics=True, axis=0)
        plt.close()

        numPix = 100
        deltaPix_source = 0.01
        f, ax = plt.subplots(1, 1, figsize=(4, 4))
        lensPlot.error_map_source_plot(
            ax=ax, numPix=numPix, deltaPix_source=deltaPix_source, with_caustics=True
        )
        plt.close()

        f, ax = plt.subplots(1, 1, figsize=(4, 4))
        lensPlot.absolute_residual_plot(ax=ax)
        plt.close()

        f, ax = plt.subplots(1, 1, figsize=(4, 4))
        lensPlot.plot_extinction_map(ax=ax)
        plt.close()

        f, ax = plt.subplots(1, 1, figsize=(4, 4))
        kwargs_plot = {"index_macromodel": [0]}
        lensPlot.substructure_plot(ax=ax, **kwargs_plot)
        plt.close()

        f, ax = plt.subplots(1, 1, figsize=(4, 4))
        kwargs_plot = {"index_macromodel": [0], "with_critical_curves": True}
        lensPlot.substructure_plot(ax=ax, **kwargs_plot)
        plt.close()

        f, ax = plt.subplots(1, 1, figsize=(4, 4))
        kwargs_plot = {"index_macromodel": [0], "with_critical_curves": True}
        lensPlot_multiplane.substructure_plot(ax=ax, **kwargs_plot)
        plt.close()

    def test_source_plot(self):
        multi_band_list = [[self.kwargs_data, self.kwargs_psf, self.kwargs_numerics]]
        lensPlot = ModelPlot(
            multi_band_list,
            self.kwargs_model,
            self.kwargs_params,
            arrow_size=0.02,
            cmap_string="gist_heat",
            fast_caustic=False,
        )

        f, ax = plt.subplots(1, 1, figsize=(4, 4))
        ax = lensPlot.source_plot(
            ax=ax,
            numPix=10,
            deltaPix_source=0.1,
            v_min=None,
            v_max=None,
            with_caustics=True,
            caustic_color="yellow",
            fsize=15,
            plot_scale="linear",
        )
        plt.close()

    def test_source(self):
        multi_band_list = [[self.kwargs_data, self.kwargs_psf, self.kwargs_numerics]]
        lensPlot = ModelPlot(
            multi_band_list,
            self.kwargs_model,
            self.kwargs_params,
            arrow_size=0.02,
            cmap_string="gist_heat",
        )
        source, coords_source = lensPlot.source(
            band_index=0, numPix=10, deltaPix=0.1, image_orientation=True
        )
        assert len(source) == 10

        source, coords_source = lensPlot.source(
            band_index=0, numPix=10, deltaPix=0.1, image_orientation=False
        )
        assert len(source) == 10

        source, coords_source = lensPlot.source(
            band_index=0, numPix=10, deltaPix=0.1, center=[0, 0]
        )
        assert len(source) == 10

    def test_joint_linear(self):
        multi_band_list = [
            [self.kwargs_data, self.kwargs_psf, self.kwargs_numerics],
            [self.kwargs_data, self.kwargs_psf, self.kwargs_numerics],
        ]
        lensPlot = ModelPlot(
            multi_band_list,
            self.kwargs_model,
            self.kwargs_params,
            arrow_size=0.02,
            cmap_string="gist_heat",
            multi_band_type="joint-linear",
            bands_compute=[True, False],
        )

        f, ax = plt.subplots(1, 1, figsize=(4, 4))
        ax = lensPlot.data_plot(
            ax=ax,
            numPix=10,
            deltaPix_source=0.1,
            v_min=None,
            v_max=None,
            with_caustics=False,
            caustic_color="yellow",
            fsize=15,
            plot_scale="linear",
        )
        plt.close()

        f, ax = plt.subplots(1, 1, figsize=(4, 4))
        ax = lensPlot.model_plot(
            ax=ax,
            numPix=10,
            deltaPix_source=0.1,
            v_min=None,
            v_max=None,
            with_caustics=False,
            caustic_color="yellow",
            fsize=15,
            plot_scale="linear",
        )
        plt.close()

        f, ax = plt.subplots(1, 1, figsize=(4, 4))
        ax = lensPlot.convergence_plot(
            ax=ax,
            numPix=10,
            deltaPix_source=0.1,
            v_min=None,
            v_max=None,
            with_caustics=False,
            caustic_color="yellow",
            fsize=15,
            plot_scale="linear",
        )
        plt.close()
        f, ax = plt.subplots(1, 1, figsize=(4, 4))
        ax = lensPlot.normalized_residual_plot(ax=ax)
        plt.close()
        f, ax = plt.subplots(1, 1, figsize=(4, 4))
        ax = lensPlot.magnification_plot(ax=ax)
        plt.close()
        f, ax = plt.subplots(1, 1, figsize=(4, 4))
        ax = lensPlot.decomposition_plot(ax=ax)
        plt.close()

    def test_reconstruction_all_bands(self):
        multi_band_list = [
            [self.kwargs_data, self.kwargs_psf, self.kwargs_numerics],
            [self.kwargs_data, self.kwargs_psf, self.kwargs_numerics],
        ]
        lensPlot = ModelPlot(
            multi_band_list,
            self.kwargs_model,
            self.kwargs_params,
            arrow_size=0.02,
            cmap_string="gist_heat",
            multi_band_type="joint-linear",
            bands_compute=[True, True],
        )
        f, axes = lensPlot.reconstruction_all_bands()
        assert len(axes) == 2
        assert len(axes[0]) == 3
        plt.close()

        multi_band_list = [[self.kwargs_data, self.kwargs_psf, self.kwargs_numerics]]
        lensPlot = ModelPlot(
            multi_band_list,
            self.kwargs_model,
            self.kwargs_params,
            arrow_size=0.02,
            cmap_string="gist_heat",
            multi_band_type="joint-linear",
            bands_compute=[True],
        )
        f, axes = lensPlot.reconstruction_all_bands()
        assert len(axes) == 1
        assert len(axes[0]) == 3
        plt.close()

    def test_check_solver_error(self):
        bool = check_solver_error(image=np.array([0, 0]))
        assert bool

        bool = check_solver_error(image=np.array([0, 0.1]))
        assert bool == 0

    def test_no_linear_solver(self):
        kwargs_data = sim_util.data_configure_simple(
            numPix=10, deltaPix=1, background_rms=1, exposure_time=1
        )
        # kwargs_data['image_data'] = np.zeros((10, 10))
        kwargs_model = {"source_light_model_list": ["GAUSSIAN"]}
        kwargs_params = {
            "kwargs_lens": [],
            "kwargs_source": [{"amp": 2, "sigma": 1, "center_x": 0, "center_y": 0}],
            "kwargs_ps": [],
            "kwargs_lens_light": [],
        }
        lensPlot = ModelPlot(
            multi_band_list=[[kwargs_data, {"psf_type": "NONE"}, {}]],
            kwargs_model=kwargs_model,
            kwargs_params=kwargs_params,
            bands_compute=[True],
            arrow_size=0.02,
            cmap_string="gist_heat",
            linear_solver=False,
        )
        lensPlot.plot_main(with_caustics=True)
        plt.close()
        assert kwargs_params["kwargs_source"][0]["amp"] == 2


class TestRaise(unittest.TestCase):
    def test_raise(self):
        with self.assertRaises(ValueError):
            kwargs_data = sim_util.data_configure_simple(
                numPix=10, deltaPix=1, background_rms=1
            )
            # kwargs_data['image_data'] = np.zeros((10, 10))
            kwargs_model = {"source_light_model_list": ["GAUSSIAN"]}
            kwargs_params = {
                "kwargs_lens": [],
                "kwargs_source": [{"amp": 1, "sigma": 1, "center_x": 0, "center_y": 0}],
                "kwargs_ps": [],
                "kwargs_lens_light": [],
            }
            lensPlot = ModelPlot(
                multi_band_list=[[kwargs_data, {"psf_type": "NONE"}, {}]],
                kwargs_model=kwargs_model,
                kwargs_params=kwargs_params,
                arrow_size=0.02,
                cmap_string="gist_heat",
            )
        with self.assertRaises(ValueError):
            kwargs_data = sim_util.data_configure_simple(
                numPix=10, deltaPix=1, background_rms=1
            )
            # kwargs_data['image_data'] = np.zeros((10, 10))
            kwargs_model = {"source_light_model_list": ["GAUSSIAN"]}
            kwargs_params = {
                "kwargs_lens": [],
                "kwargs_source": [{"amp": 1, "sigma": 1, "center_x": 0, "center_y": 0}],
                "kwargs_ps": [],
                "kwargs_lens_light": [],
            }
            lensPlot = ModelPlot(
                multi_band_list=[[kwargs_data, {}, {}]],
                kwargs_model=kwargs_model,
                kwargs_params=kwargs_params,
                arrow_size=0.02,
                cmap_string="gist_heat",
            )
            f, ax = plt.subplots(1, 1, figsize=(4, 4))
            ax = lensPlot.source_plot(
                ax=ax,
                numPix=10,
                deltaPix_source=0.1,
                v_min=None,
                v_max=None,
                with_caustics=False,
                caustic_color="yellow",
                fsize=15,
                plot_scale="bad",
            )
            plt.close()
        with self.assertRaises(ValueError):
            kwargs_data = sim_util.data_configure_simple(
                numPix=10, deltaPix=1, background_rms=1
            )
            # kwargs_data['image_data'] = np.zeros((10, 10))
            kwargs_model = {"source_light_model_list": ["GAUSSIAN"]}
            kwargs_params = {
                "kwargs_lens": [],
                "kwargs_source": [{"amp": 1, "sigma": 1, "center_x": 0, "center_y": 0}],
                "kwargs_ps": [],
                "kwargs_lens_light": [],
            }
            lensPlot = ModelPlot(
                multi_band_list=[[kwargs_data, {"psf_type": "NONE"}, {}]],
                kwargs_model=kwargs_model,
                kwargs_params=kwargs_params,
                bands_compute=[False],
                arrow_size=0.02,
                cmap_string="gist_heat",
            )
            lensPlot._select_band(band_index=0)

        with self.assertRaises(ValueError):
            kwargs_data = sim_util.data_configure_simple(
                numPix=10, deltaPix=1, background_rms=1, exposure_time=1
            )
            # kwargs_data['image_data'] = np.zeros((10, 10))
            kwargs_model = {"source_light_model_list": ["GAUSSIAN"]}
            kwargs_params = {
                "kwargs_lens": [],
                "kwargs_source": [{"amp": 1, "sigma": 1, "center_x": 0, "center_y": 0}],
                "kwargs_ps": [],
                "kwargs_lens_light": [],
            }
            lensPlot = ModelPlot(
                multi_band_list=[[kwargs_data, {"psf_type": "NONE"}, {}]],
                kwargs_model=kwargs_model,
                kwargs_params=kwargs_params,
                bands_compute=[True],
                arrow_size=0.02,
                cmap_string="gist_heat",
            )

            f, ax = plt.subplots(1, 1, figsize=(4, 4))
            ax = lensPlot.source_plot(
                ax=ax,
                numPix=10,
                deltaPix_source=0.1,
                v_min=None,
                v_max=None,
                with_caustics=False,
                caustic_color="yellow",
                fsize=15,
                plot_scale="wrong",
            )
            plt.close()
        with self.assertRaises(ValueError):
            # test whether linear_solver=False returns raise when having two bands
            kwargs_data = sim_util.data_configure_simple(
                numPix=10, deltaPix=1, background_rms=1, exposure_time=1
            )
            # kwargs_data['image_data'] = np.zeros((10, 10))
            kwargs_model = {"source_light_model_list": ["GAUSSIAN"]}
            kwargs_params = {
                "kwargs_lens": [],
                "kwargs_source": [{"amp": 2, "sigma": 1, "center_x": 0, "center_y": 0}],
                "kwargs_ps": [],
                "kwargs_lens_light": [],
            }
            lensPlot = ModelPlot(
                multi_band_list=[
                    [kwargs_data, {"psf_type": "NONE"}, {}],
                    [kwargs_data, {"psf_type": "NONE"}, {}],
                ],
                kwargs_model=kwargs_model,
                kwargs_params=kwargs_params,
                bands_compute=[True],
                arrow_size=0.02,
                cmap_string="gist_heat",
                linear_solver=False,
            )


def test_interferometry_natwt_Model_Plot_linear_solver():
    # Test no errors are raised in the Model Plot linear solver for 'interferometry_natwt' likelihood function.
    try:
        kwargs_data = sim_util.data_configure_simple(
            numPix=10, deltaPix=1, background_rms=1, exposure_time=1
        )
        kwargs_data["likelihood_method"] = "interferometry_natwt"
        kwargs_model = {"source_light_model_list": ["GAUSSIAN"]}
        kwargs_params = {
            "kwargs_lens": [],
            "kwargs_source": [{"amp": 2, "sigma": 1, "center_x": 0, "center_y": 0}],
            "kwargs_ps": [],
            "kwargs_lens_light": [],
        }
        lensPlot = ModelPlot(
            multi_band_list=[[kwargs_data, {"psf_type": "NONE"}, {}]],
            kwargs_model=kwargs_model,
            kwargs_params=kwargs_params,
            bands_compute=[True],
            arrow_size=0.02,
            cmap_string="gist_heat",
            linear_solver=True,
        )
    except:
        pytest.fail(
            "Errors are raised in the Model Plot linear solver for the 'interferometric_natwt' likelihood method, which is not expected."
        )


if __name__ == "__main__":
    pytest.main()
